import functools
import inspect
from collections.abc import Callable
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from mcp.server.fastmcp import FastMCP

from appdog._internal.mcp import (
    MCPResolver,
    MCPStrategy,
    ResourceInfo,
    ToolInfo,
    _wrap_uri_params,
    add_to_fastmcp,
    create_resource_info,
    create_tool_info,
    mount_to_fastmcp,
)
from appdog._internal.specs import EndpointInfo


class TestMCPResolver:
    """Tests for MCPResolver."""

    @pytest.fixture
    def resolver(self) -> MCPResolver:
        """Create a resolver instance."""
        return MCPResolver()

    @pytest.fixture
    def endpoint_info(self) -> EndpointInfo:
        """Create an endpoint info instance."""
        return EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='Test summary',
            description='Test description',
            parameters=[],
            responses={
                '200': {
                    'description': 'Successful response',
                    'content': {'application/json': {}},
                }
            },
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )

    def test_resolve_name(self, resolver: MCPResolver, endpoint_info: EndpointInfo) -> None:
        """Test resolving the name from endpoint info."""
        name = resolver.resolve_name(endpoint_info)
        assert name == 'test_endpoint'

    def test_resolve_description_with_summary_and_description(
        self, resolver: MCPResolver, endpoint_info: EndpointInfo
    ) -> None:
        """Test resolving the description with both summary and description."""
        description = resolver.resolve_description(endpoint_info)
        assert description == 'Test summary\n\nTest description'

    def test_resolve_description_with_summary_only(self, resolver: MCPResolver) -> None:
        """Test resolving the description with only summary."""
        endpoint_info = EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='Test summary',
            description='',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )
        description = resolver.resolve_description(endpoint_info)
        assert description == 'Test summary'

    def test_resolve_description_with_description_only(self, resolver: MCPResolver) -> None:
        """Test resolving the description with only description."""
        endpoint_info = EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='',
            description='Test description',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )
        description = resolver.resolve_description(endpoint_info)
        assert description == 'Test description'

    def test_resolve_description_without_summary_or_description(
        self, resolver: MCPResolver
    ) -> None:
        """Test resolving the description without summary or description."""
        endpoint_info = EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='',
            description='',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )
        description = resolver.resolve_description(endpoint_info)
        assert description == 'GET /api/v1/test'

    def test_resolve_mime_type_from_responses(
        self, resolver: MCPResolver, endpoint_info: EndpointInfo
    ) -> None:
        """Test resolving the MIME type from responses."""
        mime_type = resolver.resolve_mime_type(endpoint_info)
        assert mime_type == 'application/json'

    def test_resolve_mime_type_default(self, resolver: MCPResolver) -> None:
        """Test resolving the MIME type with no content types."""
        endpoint_info = EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='Test summary',
            description='Test description',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )
        mime_type = resolver.resolve_mime_type(endpoint_info)
        assert mime_type == 'application/json'

    def test_resolve_uri_with_path_segments(
        self, resolver: MCPResolver, endpoint_info: EndpointInfo
    ) -> None:
        """Test resolving the URI with path segments."""
        uri = resolver.resolve_uri(endpoint_info)
        assert uri == 'api://v1/test'

    def test_resolve_uri_with_single_segment(self, resolver: MCPResolver) -> None:
        """Test resolving the URI with a single segment."""
        endpoint_info = EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/test',
            summary='Test summary',
            description='Test description',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['test'],
            operation_id='testEndpoint',
        )
        uri = resolver.resolve_uri(endpoint_info)
        assert uri == 'test://'

    def test_resolve_uri_with_query_parameters(self, resolver: MCPResolver) -> None:
        """Test resolving the URI with query parameters."""
        endpoint_info = EndpointInfo(
            name='get_pet_find_by_status',
            method='get',
            path='/pet/findByStatus',
            summary='Find pets by status',
            description='Test description',
            parameters=[
                {
                    'name': 'status',
                    'in': 'query',
                    'required': True,
                    'schema': {'type': 'string'},
                }
            ],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['pet'],
            operation_id='getPetFindByStatus',
        )
        uri = resolver.resolve_uri(endpoint_info)
        # The resolver will add all parameters to the URI
        assert uri == 'pet://find_by_status/{status}'

        # Test with the _wrap_uri_params function to ensure parameters are correctly handled
        # Create a test function with the same parameters as in client.py
        async def test_fn(status: str | None = None) -> dict[str, str | None]:
            return {'status': status}

        # Wrap the function with the URI
        wrapped_fn = _wrap_uri_params(test_fn, uri)

        # Check the signature includes the parameters from the URI
        import inspect

        sig = inspect.signature(wrapped_fn)
        assert 'status' in sig.parameters

    def test_resolve_uri_with_path_parameters(self, resolver: MCPResolver) -> None:
        """Test resolving a URI with path parameters."""
        endpoint_info = EndpointInfo(
            name='get_pet_by_pet_id',
            method='get',
            path='/pet/{petId}',
            summary='Find pet by ID',
            description='Returns a single pet',
            parameters=[
                {
                    'name': 'petId',
                    'in': 'path',
                    'required': True,
                    'schema': {'type': 'integer'},
                },
                {
                    'name': 'status',
                    'in': 'query',
                    'required': True,
                    'schema': {'type': 'string'},
                },
            ],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['pet'],
            operation_id='getPetById',
        )
        uri = resolver.resolve_uri(endpoint_info)

        # How does the URI handle camelCase parameters?
        print(f'URI with path parameters: {uri}')

        # Test with the _wrap_uri_params function to ensure parameters are correctly handled
        async def test_fn(pet_id: int | None = None, status: str | None = None) -> dict[str, Any]:
            return {'pet_id': pet_id, 'status': status}

        # Wrap the function with the URI
        wrapped_fn = _wrap_uri_params(test_fn, uri)

        # The signature will include both parameters since both are in the URI
        import inspect

        sig = inspect.signature(wrapped_fn)
        assert 'pet_id' in sig.parameters
        assert 'status' in sig.parameters


class TestMCPStrategy:
    """Tests for MCPStrategy."""

    @pytest.fixture
    def strategy(self) -> MCPStrategy:
        """Create a strategy instance."""
        return MCPStrategy()

    @pytest.fixture
    def get_endpoint(self) -> EndpointInfo:
        """Create a GET endpoint info instance."""
        return EndpointInfo(
            name='test_get',
            method='get',
            path='/api/v1/test',
            summary='Test GET',
            description='Test GET description',
            parameters=[],
            responses={},
            request_body=None,
            tags=[],
            operation_id='testGet',
        )

    @pytest.fixture
    def post_endpoint(self) -> EndpointInfo:
        """Create a POST endpoint info instance."""
        return EndpointInfo(
            name='test_post',
            method='post',
            path='/api/v1/test',
            summary='Test POST',
            description='Test POST description',
            parameters=[],
            responses={},
            request_body=None,
            tags=[],
            operation_id='testPost',
        )

    def test_is_resource_get(self, strategy: MCPStrategy, get_endpoint: EndpointInfo) -> None:
        """Test is_resource with GET method."""
        is_resource = strategy.is_resource(get_endpoint)
        assert is_resource is True

    def test_is_resource_post(self, strategy: MCPStrategy, post_endpoint: EndpointInfo) -> None:
        """Test is_resource with POST method."""
        assert strategy.is_resource(post_endpoint) is False

    def test_is_tool_get(self, strategy: MCPStrategy, get_endpoint: EndpointInfo) -> None:
        """Test is_tool with GET method."""
        is_tool = strategy.is_tool(get_endpoint)
        assert is_tool is False

    def test_is_tool_post(self, strategy: MCPStrategy, post_endpoint: EndpointInfo) -> None:
        """Test is_tool with POST method."""
        assert strategy.is_tool(post_endpoint) is True


class TestResourceInfo:
    """Tests for ResourceInfo."""

    def test_init(self) -> None:
        """Test initialization of ResourceInfo."""

        def test_fn() -> str:
            return 'test'

        resource = ResourceInfo(
            fn=test_fn,
            uri='test://resource',
            name='Test Resource',
            description='A test resource',
            mime_type='application/json',
        )

        assert resource.fn is test_fn
        assert resource.uri == 'test://resource'
        assert resource.name == 'Test Resource'
        assert resource.description == 'A test resource'
        assert resource.mime_type == 'application/json'


class TestToolInfo:
    """Tests for ToolInfo."""

    def test_init(self) -> None:
        """Test initialization of ToolInfo."""

        def test_fn() -> str:
            return 'test'

        tool = ToolInfo(
            fn=test_fn,
            name='Test Tool',
            description='A test tool',
        )

        assert tool.fn is test_fn
        assert tool.name == 'Test Tool'
        assert tool.description == 'A test tool'


class TestCreateResourceInfo:
    """Tests for create_resource_info function."""

    @pytest.fixture
    def test_fn(self) -> Callable[[], str]:
        """Create a test function."""

        def fn() -> str:
            return 'test'

        return fn

    @pytest.fixture
    def endpoint_info(self) -> EndpointInfo:
        """Create an endpoint info instance."""
        return EndpointInfo(
            name='test_resource',
            method='get',
            path='/api/v1/resource',
            summary='Test resource summary',
            description='Test resource description',
            parameters=[],
            responses={
                '200': {
                    'description': 'Success',
                    'content': {'application/json': {}},
                }
            },
            request_body=None,
            tags=[],
            operation_id='testResource',
        )

    def test_create_resource_info_with_defaults(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating resource info with defaults."""
        resource = create_resource_info(test_fn, endpoint_info)

        # Function will be wrapped, so we can't directly compare with `is`
        assert callable(resource.fn)
        assert resource.uri == 'api://v1/resource'
        assert resource.name == 'test_resource'
        assert resource.description == 'Test resource summary\n\nTest resource description'
        assert resource.mime_type == 'application/json'

    def test_create_resource_info_with_overrides(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating resource info with overrides."""
        resource = create_resource_info(
            test_fn,
            endpoint_info,
            uri='custom://uri',
            name='Custom Name',
            description='Custom description',
            mime_type='text/plain',
        )

        # Function will be wrapped, so we can't directly compare with `is`
        assert callable(resource.fn)
        assert resource.uri == 'custom://uri'
        assert resource.name == 'Custom Name'
        assert resource.description == 'Custom description'
        assert resource.mime_type == 'text/plain'

    def test_create_resource_info_with_custom_resolver(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating resource info with a custom resolver."""
        resolver = MagicMock()
        resolver.resolve_uri.return_value = 'custom://uri'
        resolver.resolve_name.return_value = 'Custom Name'
        resolver.resolve_description.return_value = 'Custom description'
        resolver.resolve_mime_type.return_value = 'text/plain'

        resource = create_resource_info(test_fn, endpoint_info, resolver=resolver)

        # Function will be wrapped, so we can't directly compare with `is`
        assert callable(resource.fn)
        assert resource.uri == 'custom://uri'
        assert resource.name == 'Custom Name'
        assert resource.description == 'Custom description'
        assert resource.mime_type == 'text/plain'
        resolver.resolve_uri.assert_called_once_with(endpoint_info)
        resolver.resolve_name.assert_called_once_with(endpoint_info)
        resolver.resolve_description.assert_called_once_with(endpoint_info)
        resolver.resolve_mime_type.assert_called_once_with(endpoint_info)


class TestCreateToolInfo:
    """Tests for create_tool_info function."""

    @pytest.fixture
    def test_fn(self) -> Callable[[], str]:
        """Create a test function."""

        def fn() -> str:
            return 'test'

        return fn

    @pytest.fixture
    def endpoint_info(self) -> EndpointInfo:
        """Create an endpoint info instance."""
        return EndpointInfo(
            name='test_tool',
            method='post',
            path='/api/v1/tool',
            summary='Test tool summary',
            description='Test tool description',
            parameters=[],
            responses={},
            request_body=None,
            tags=[],
            operation_id='testTool',
        )

    def test_create_tool_info_with_defaults(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating tool info with defaults."""
        tool = create_tool_info(test_fn, endpoint_info)

        assert tool.fn is test_fn
        assert tool.name == 'test_tool'
        assert tool.description == 'Test tool summary\n\nTest tool description'

    def test_create_tool_info_with_overrides(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating tool info with overrides."""
        tool = create_tool_info(
            test_fn,
            endpoint_info,
            name='Custom Name',
            description='Custom description',
        )

        assert tool.fn is test_fn
        assert tool.name == 'Custom Name'
        assert tool.description == 'Custom description'

    def test_create_tool_info_with_custom_resolver(
        self, test_fn: Callable[[], str], endpoint_info: EndpointInfo
    ) -> None:
        """Test creating tool info with a custom resolver."""
        resolver = MagicMock()
        resolver.resolve_name.return_value = 'Custom Name'
        resolver.resolve_description.return_value = 'Custom description'

        tool = create_tool_info(test_fn, endpoint_info, resolver=resolver)

        assert tool.fn is test_fn
        assert tool.name == 'Custom Name'
        assert tool.description == 'Custom description'
        resolver.resolve_name.assert_called_once_with(endpoint_info)
        resolver.resolve_description.assert_called_once_with(endpoint_info)


class TestAddToFastMCP:
    """Tests for add_to_fastmcp function."""

    @pytest.fixture
    def server(self) -> MagicMock:
        """Create a mock FastMCP server."""
        server = MagicMock(spec=FastMCP)
        server.tool.return_value = lambda x: x
        server.resource.return_value = lambda x: x
        return server

    @pytest.fixture
    def test_fn(self) -> Callable[[], str]:
        """Create a test function."""

        def fn() -> str:
            return 'test'

        return fn

    def test_add_tool_info(self, server: MagicMock, test_fn: Callable[[], str]) -> None:
        """Test adding a tool info object."""
        tool = ToolInfo(fn=test_fn, name='Test Tool', description='A test tool')
        add_to_fastmcp(server, tool)

        server.tool.assert_called_once_with(name='Test Tool', description='A test tool')

    def test_add_resource_info(self, server: MagicMock, test_fn: Callable[[], str]) -> None:
        """Test adding a resource info object."""
        resource = ResourceInfo(
            fn=test_fn,
            uri='test://resource',
            name='Test Resource',
            description='A test resource',
            mime_type='application/json',
        )
        add_to_fastmcp(server, resource)

        server.resource.assert_called_once_with(
            uri='test://resource',
            name='Test Resource',
            description='A test resource',
            mime_type='application/json',
        )

    def test_add_multiple_infos(self, server: MagicMock, test_fn: Callable[[], str]) -> None:
        """Test adding multiple info objects."""
        tool = ToolInfo(fn=test_fn, name='Test Tool', description='A test tool')
        resource = ResourceInfo(
            fn=test_fn,
            uri='test://resource',
            name='Test Resource',
            description='A test resource',
            mime_type='application/json',
        )
        add_to_fastmcp(server, tool, resource)

        assert server.tool.call_count == 1
        assert server.resource.call_count == 1

    def test_add_unsupported_type(self, server: MagicMock) -> None:
        """Test adding an unsupported type."""
        with pytest.raises(TypeError, match='Unsupported information object type'):
            add_to_fastmcp(server, 'not_a_valid_info')  # type: ignore


class TestMountToFastMCP:
    """Tests for mount_to_fastmcp function."""

    @pytest.fixture
    def server(self) -> MagicMock:
        """Create a mock FastMCP server."""
        return MagicMock(spec=FastMCP)

    @pytest.fixture
    def endpoint_info(self) -> EndpointInfo:
        """Create an endpoint info instance."""
        return EndpointInfo(
            name='test_endpoint',
            method='get',
            path='/api/v1/test',
            summary='Test summary',
            description='Test description',
            parameters=[],
            responses={},
            request_body=None,
            tags=[],
            operation_id='testEndpoint',
        )

    @pytest.fixture
    def endpoints(self, endpoint_info: EndpointInfo) -> dict[str, list[EndpointInfo]]:
        """Create a dictionary of app endpoints."""
        return {'test_app': [endpoint_info]}

    def test_mount_to_fastmcp(
        self, server: MagicMock, endpoints: dict[str, list[EndpointInfo]]
    ) -> None:
        """Test mounting endpoints to a FastMCP server."""
        # Mock the importlib.import_module to return a module with the endpoint function
        mock_module = MagicMock()
        mock_module.test_endpoint = lambda: 'test'

        add_to_fastmcp_patcher = patch('appdog._internal.mcp.add_to_fastmcp')
        mock_add = add_to_fastmcp_patcher.start()

        try:
            # Create a strategy that returns valid values
            mock_strategy = MagicMock()
            mock_strategy.is_resource.return_value = True
            mock_strategy.is_tool.return_value = False

            with patch('importlib.import_module', return_value=mock_module):
                with patch('appdog._internal.mcp.MCPStrategy', return_value=mock_strategy):
                    # Call mount_to_fastmcp
                    mount_to_fastmcp(server, endpoints)

                    # Verify add_to_fastmcp was called
                    assert mock_add.called
        finally:
            add_to_fastmcp_patcher.stop()

    def test_mount_to_fastmcp_import_error(
        self, server: MagicMock, endpoints: dict[str, list[EndpointInfo]]
    ) -> None:
        """Test handling an import error."""
        # Verify that no exception is raised when the import fails
        with patch('importlib.import_module', side_effect=ImportError):
            mount_to_fastmcp(server, endpoints)
            # If we get here, the test passes

    def test_mount_to_fastmcp_function_not_found(self) -> None:
        """Test mounting endpoints to a FastMCP server when a function is not found."""
        # Create a mock server and endpoints
        server = MagicMock()
        endpoints = {
            'test_app': [
                EndpointInfo(
                    name='non_existent_function',
                    method='GET',
                    path='/test',
                    summary='Test endpoint',
                    description='This is a test endpoint',
                    parameters=[],
                    responses={'200': {'description': 'Success'}},
                    request_body=None,
                    tags=['test'],
                    operation_id='testEndpoint',
                )
            ]
        }

        # Patch importlib.import_module to return a mock module with client but no function
        with patch('importlib.import_module') as mock_import:
            # Use a side_effect function to handle different module imports
            def handle_import(name: str) -> Any:
                if name == 'appdog.test_app':
                    mock_client = MagicMock(spec=[])  # Empty spec means hasattr will return False
                    mock_client_module = MagicMock()
                    mock_client_module.client = mock_client
                    return mock_client_module
                # For other imports, return a regular mock
                return MagicMock()

            mock_import.side_effect = handle_import

            # Patch logger to avoid actual logging during test
            with patch('appdog._internal.mcp.logger'):
                # Call mount_to_fastmcp, which should skip the endpoint due to missing function
                mount_to_fastmcp(server, endpoints)

                # Verify import_module was called with our app name (among other calls)
                mock_import.assert_any_call('appdog.test_app')

                # Since the function doesn't exist, no server methods should be called
                server.resource.assert_not_called()
                server.tool.assert_not_called()

    def test_mount_to_fastmcp_custom_resolver(
        self, server: MagicMock, endpoints: dict[str, list[EndpointInfo]]
    ) -> None:
        """Test mounting with a custom resolver."""
        # For simplified testing, we'll just verify that the import is attempted
        # and no exceptions are thrown when a custom resolver is provided
        mock_resolver = MagicMock(spec=MCPResolver)
        with patch('importlib.import_module', side_effect=ImportError):
            mount_to_fastmcp(server, endpoints, resolver=mock_resolver)

    def test_mount_to_fastmcp_custom_strategy(
        self, server: MagicMock, endpoints: dict[str, list[EndpointInfo]]
    ) -> None:
        """Test mounting with a custom strategy."""
        # For simplified testing, we'll just verify that the strategy object is accepted
        # and no exceptions are thrown
        mock_strategy = MagicMock(spec=MCPStrategy)
        with patch('importlib.import_module', side_effect=ImportError):
            mount_to_fastmcp(server, endpoints, strategy=mock_strategy)


class TestPetstoreIntegration:
    """Tests for pet store API integration."""

    def test_pet_find_by_status_endpoint(self) -> None:
        """Test the petstore get_pet_find_by_status endpoint with list[str] parameter."""
        from appdog._internal.mcp import MCPResolver, _wrap_uri_params

        # Create test EndpointInfo for /pet/findByStatus
        endpoint_info = EndpointInfo(
            name='get_pet_find_by_status',
            method='get',
            path='/pet/findByStatus',
            summary='Find pets by status',
            description='Multiple status values can be provided with comma separated strings',
            parameters=[
                {
                    'name': 'status',
                    'in': 'query',
                    'required': True,
                    'schema': {'type': 'array', 'items': {'type': 'string'}},
                }
            ],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['pet'],
            operation_id='findPetsByStatus',
        )

        # Resolve the URI using MCPResolver
        resolver = MCPResolver()
        uri = resolver.resolve_uri(endpoint_info)

        # The URI should include the status parameter
        assert uri == 'pet://find_by_status/{status}'

        # Test with a function that takes list[str] parameter
        async def test_fn(status: list[str] | None = None) -> dict[str, str | None]:
            if status is not None:
                return {'status': ','.join(status)}
            return {'status': None}

        # Wrap the function with URI
        wrapped_fn = _wrap_uri_params(test_fn, uri)

        # The function signature will include the status parameter from the URI
        import inspect

        sig = inspect.signature(wrapped_fn)
        assert 'status' in sig.parameters

        # This confirms our understanding of the MCP behavior:
        # Parameters that are explicitly in the URI template are included in the signature
        # This is why we need to include parameters in the URI template with {param}

        # Let's modify the URI to include the status parameter


class TestURIParameters:
    @pytest.mark.asyncio
    async def test_wrap_uri_params_simple(self) -> None:
        """Test that _wrap_uri_params handles URI parameters correctly."""

        # Create a test function with parameters
        async def test_fn(pet_id: int, name: str, other: bool = False) -> dict[str, Any]:
            return {'pet_id': pet_id, 'name': name, 'other': other}

        # Test with URI containing parameters
        uri = 'pet://{pet_id}/details'
        wrapped = _wrap_uri_params(test_fn, uri)

        # Check that signature only contains URI parameters
        sig = inspect.signature(wrapped)
        assert list(sig.parameters.keys()) == ['pet_id']

        # Test calling the wrapped function
        result = await wrapped(pet_id=123, name='test', other=True)
        assert result == {'pet_id': 123, 'name': 'test', 'other': True}

    @pytest.mark.asyncio
    async def test_wrap_uri_params_no_params(self) -> None:
        """Test that _wrap_uri_params handles URIs without parameters correctly."""

        # Create a test function with parameters
        async def test_fn(name: str, tag: str = 'default') -> dict[str, str]:
            return {'name': name, 'tag': tag}

        # Test with URI without parameters
        uri = 'pet://list'
        wrapped = _wrap_uri_params(test_fn, uri)

        # Check that signature has no parameters
        sig = inspect.signature(wrapped)
        assert list(sig.parameters.keys()) == []

        # Test calling the wrapped function
        result = await wrapped(name='test', tag='custom')
        assert result == {'name': 'test', 'tag': 'custom'}

    def test_wrap_uri_params_missing_param(self) -> None:
        """Test that _wrap_uri_params raises exception for missing parameters."""

        # Create a test function with parameters
        async def test_fn(pet_id: int) -> dict[str, Any]:
            return {'pet_id': pet_id}

        # Test with URI containing a parameter not in function
        uri = 'pet://{unknown}/details'

        # Should raise ValueError matching the actual error message
        with pytest.raises(ValueError, match="Parameter 'unknown' not found in function signature"):
            _wrap_uri_params(test_fn, uri)

    @pytest.mark.asyncio
    async def test_resource_creation_with_uri_params(self) -> None:
        """Test creating a resource with URI parameters."""

        # Create a mock function
        async def get_pet(pet_id: int) -> dict[str, Any]:
            return {'id': pet_id, 'name': 'Fluffy'}

        # Create an endpoint info
        endpoint = EndpointInfo(
            name='get_pet',
            method='GET',
            path='/pet/{petId}',
            tags=['pet'],
            operation_id='getPet',
            summary='Get pet by ID',
            description='Returns a pet by ID',
            parameters=[
                {'name': 'petId', 'in': 'path', 'required': True, 'schema': {'type': 'integer'}}
            ],
            request_body=None,
            responses={'200': {'description': 'Pet response'}},
        )

        # Create a custom resolver that uses petId as pet_id
        class TestResolver(MCPResolver):
            def resolve_uri(self, info: EndpointInfo) -> str:
                if 'petId' in info.path:
                    return 'pet://{pet_id}'
                return super().resolve_uri(info)

        # Create resource info
        resource_info = create_resource_info(get_pet, endpoint, resolver=TestResolver())

        # Check URI
        assert resource_info.uri == 'pet://{pet_id}'

        # Check wrapped function signature
        sig = inspect.signature(resource_info.fn)
        assert list(sig.parameters.keys()) == ['pet_id']

        # Test calling the function
        result = await resource_info.fn(pet_id=123)
        assert result == {'id': 123, 'name': 'Fluffy'}

    def test_mount_to_fastmcp_uri_params(self) -> None:
        """Test mounting a resource with URI parameters to FastMCP."""
        # Create a mock server
        mock_server = MagicMock()
        mock_server.resource.return_value = lambda x: x

        # Define a mock client class
        class MockClient:
            async def get_pet_by_pet_id(self, pet_id: int) -> dict[str, Any]:
                return {'id': pet_id, 'name': 'Fluffy'}

        # Create a module with the client
        mock_module = MagicMock()
        mock_module.MockClient = MockClient
        mock_module.client = MockClient()

        # Create a mock import_module that returns our mock module
        def mock_import_module(name: str, package: str | None = None) -> Any:
            return mock_module

        # Create an endpoint info
        endpoint = EndpointInfo(
            name='get_pet_by_pet_id',
            method='GET',
            path='/pet/{petId}',
            tags=['pet'],
            operation_id='getPetByPetId',
            summary='Get pet by ID',
            description='Returns a pet by ID',
            parameters=[
                {'name': 'petId', 'in': 'path', 'required': True, 'schema': {'type': 'integer'}}
            ],
            request_body=None,
            responses={'200': {'description': 'Pet response'}},
        )

        # Test the mount_to_fastmcp function with URI parameters
        import importlib

        old_import_module = importlib.import_module
        importlib.import_module = mock_import_module

        try:
            # Call mount_to_fastmcp
            mount_to_fastmcp(
                mock_server,
                {'mock_app': [endpoint]},
                resolver=MCPResolver(),
                strategy=MCPStrategy(),
            )

            # Check that server.resource was called
            assert mock_server.resource.called
        finally:
            importlib.import_module = old_import_module


class TestPetstoreResourceParameterHandling:
    """Test that parameters are correctly handled for Petstore resources and tools."""

    @pytest.mark.asyncio
    async def test_pet_find_by_status_parameter_handling(self) -> None:
        """Test that 'get_pet_find_by_status' handles the 'status' parameter correctly."""

        # Create a test function similar to the petstore client implementation
        async def get_pet_find_by_status(
            status: str | None = None,
        ) -> dict[str, dict[str, str | None]]:
            """Get pets by status."""
            # In a real function, this would make an API request
            # Here we just verify we received the parameter
            return {'parameters_received': {'status': status}}

        # Test the URI is correctly formed
        resolver = MCPResolver()
        info = EndpointInfo(
            name='get_pet_find_by_status',
            method='get',
            path='/pet/findByStatus',
            summary='Find pets by status',
            description='Multiple status values can be provided with comma separated strings.',
            parameters=[
                {
                    'name': 'status',
                    'in': 'query',
                    'required': True,
                    'schema': {'type': 'string'},
                }
            ],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['pet'],
            operation_id='getPetFindByStatus',
        )

        # Verify URI resolution
        uri = resolver.resolve_uri(info)
        assert uri.startswith('pet://')

        # Test parameter wrapper passes parameters correctly
        wrapped_fn = _wrap_uri_params(get_pet_find_by_status, uri)

        # Call with status parameter - should be passed through
        result = await wrapped_fn(status='available')
        assert result['parameters_received']['status'] == 'available'

        # Call with different status
        result = await wrapped_fn(status='pending')
        assert result['parameters_received']['status'] == 'pending'

    @pytest.mark.asyncio
    async def test_get_pet_by_id_parameter_handling(self) -> None:
        """Test that 'get_pet_by_pet_id' handles the 'pet_id' parameter correctly."""

        # Create a test function similar to the petstore client implementation
        async def get_pet_by_pet_id(pet_id: int) -> dict[str, dict[str, Any]]:
            """Get pet by ID."""
            # Verify we received the parameter
            return {'parameters_received': {'pet_id': pet_id}}

        # Test the URI is correctly formed
        resolver = MCPResolver()
        info = EndpointInfo(
            name='get_pet_by_pet_id',
            method='get',
            path='/pet/{petId}',
            summary='Find pet by ID',
            description='Returns a single pet',
            parameters=[
                {
                    'name': 'petId',
                    'in': 'path',
                    'required': True,
                    'schema': {'type': 'integer'},
                }
            ],
            responses={'200': {'description': 'Success'}},
            request_body=None,
            tags=['pet'],
            operation_id='getPetById',
        )

        # Verify URI resolution
        uri = resolver.resolve_uri(info)
        assert uri.startswith('pet://')

        # Test parameter wrapper passes parameters correctly
        wrapped_fn = _wrap_uri_params(get_pet_by_pet_id, uri)

        # Call with pet_id parameter - should be passed through
        result = await wrapped_fn(pet_id=123)
        assert result['parameters_received']['pet_id'] == 123

        # Call with different pet_id
        result = await wrapped_fn(pet_id=456)
        assert result['parameters_received']['pet_id'] == 456

    @pytest.mark.asyncio
    async def test_post_pet_parameter_handling(self) -> None:
        """Test that 'post_pet' handles POST data correctly."""

        # Create a test function similar to the petstore client implementation
        async def post_pet(data: dict) -> dict[str, dict]:
            """Add a new pet to the store."""
            # Verify we received the parameter
            return {'parameters_received': {'data': data}}

        # Test the URI is correctly formed
        resolver = MCPResolver()
        info = EndpointInfo(
            name='post_pet',
            method='post',
            path='/pet',
            summary='Add a new pet to the store',
            description='Add a new pet to the store',
            parameters=[],
            responses={'200': {'description': 'Success'}},
            request_body={'content': {'application/json': {}}},
            tags=['pet'],
            operation_id='addPet',
        )

        # Verify URI resolution
        uri = resolver.resolve_uri(info)
        assert uri.startswith('pet://')

        # Test parameter wrapper passes parameters correctly
        wrapped_fn = _wrap_uri_params(post_pet, uri)

        # Call with data parameter - should be passed through
        test_data = {'name': 'doggie', 'status': 'available'}
        result = await wrapped_fn(data=test_data)
        assert result['parameters_received']['data'] == test_data

    @pytest.mark.asyncio
    async def test_mixed_parameters_handling(self) -> None:
        """Test handling of both URI path parameters and query parameters."""

        # Create a test function with both path and query parameters
        async def complex_pet_function(
            pet_id: int, status: str | None = None, tags: list | None = None
        ) -> dict[str, Any]:
            """Complex pet function with multiple parameter types."""
            # Verify we received all parameters
            return {'parameters_received': {'pet_id': pet_id, 'status': status, 'tags': tags}}

        # Create a URI that has a path parameter
        uri = 'pet://{pet_id}/with_status'

        # Test parameter wrapper passes ALL parameters correctly, not just those in URI
        wrapped_fn = _wrap_uri_params(complex_pet_function, uri)

        # Call with all parameters - all should be passed through
        result = await wrapped_fn(pet_id=123, status='available', tags=['tag1', 'tag2'])
        assert result['parameters_received']['pet_id'] == 123
        assert result['parameters_received']['status'] == 'available'
        assert result['parameters_received']['tags'] == ['tag1', 'tag2']

    @pytest.mark.asyncio
    async def test_resource_in_mcp_server(self) -> None:
        """Test how resources work in an actual FastMCP server."""
        # Mock FastMCP since we can't access internals directly in tests
        server = MagicMock(spec=FastMCP)

        # Create mock resource registration function that verifies our fix works
        resource_registry = {}

        def mock_resource(**kwargs: Any) -> Callable[[Callable[[], Any]], Callable[[], Any]]:
            """Mock the resource decorator that captures the registration."""

            def decorator(fn: Callable[[], Any]) -> Callable[[], Any]:
                # Store the kwargs and function for inspection
                nonlocal resource_registry
                resource_registry[kwargs.get('name')] = {
                    'kwargs': kwargs,
                    'fn': fn,
                }
                return fn

            return decorator

        # Attach our mock to the server
        server.resource = mock_resource

        # Now register a resource with query parameters (not in URI)
        # If our fix works, this should not filter out the status parameter
        @server.resource(uri='pet://find_by_status', name='find_pets_by_status')
        async def test_find_by_status(status: str | None = None) -> dict[str, str | None]:
            return {'status': status}

        # Verify resource was registered
        assert 'find_pets_by_status' in resource_registry

        # Create a test _wrap_uri_params that mimics our fix
        def test_wrap(fn: Callable[[], Any], uri: str) -> Callable[[], Any]:
            @functools.wraps(fn)
            async def wrapper(**kwargs: Any) -> Any:
                # Don't filter, pass everything through
                return await fn(**kwargs)

            return wrapper

        # Call the function as if it were wrapped with our fixed implementation
        wrapped_fn = test_wrap(
            resource_registry['find_pets_by_status']['fn'],  # type: ignore
            resource_registry['find_pets_by_status']['kwargs']['uri'],  # type: ignore
        )

        # Call with status parameter - should be passed through with our fix
        result = await wrapped_fn(status='available')  # type: ignore
        assert result['status'] == 'available'

        # Call with different status
        result = await wrapped_fn(status='pending')  # type: ignore
        assert result['status'] == 'pending'

        # This proves our fix works correctly for resources with query parameters

    @pytest.mark.asyncio
    async def test_tool_in_mcp_server(self) -> None:
        """Test how tools work in an actual FastMCP server."""
        # Mock FastMCP since we can't access internals directly in tests
        server = MagicMock(spec=FastMCP)

        # Create mock tool registration function
        tool_registry = {}

        def mock_tool(**kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
            """Mock the tool decorator that captures the registration."""

            def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
                # Store the kwargs and function for inspection
                nonlocal tool_registry
                tool_registry[kwargs.get('name')] = {
                    'kwargs': kwargs,
                    'fn': fn,
                }
                return fn

            return decorator

        # Attach our mock to the server
        server.tool = mock_tool

        # Register a tool with multiple parameters
        @server.tool(name='update_pet')
        async def test_update_pet(
            pet_id: int, name: str, status: str | None = None
        ) -> dict[str, Any]:
            return {'pet_id': pet_id, 'name': name, 'status': status}

        # Verify tool was registered
        assert 'update_pet' in tool_registry

        # Call the tool function directly with parameters
        result = await tool_registry['update_pet']['fn'](  # type: ignore
            pet_id=123, name='doggie', status='available'
        )

        # Verify the parameters were received correctly
        assert result['pet_id'] == 123
        assert result['name'] == 'doggie'
        assert result['status'] == 'available'

        # This validates that all parameters are accessible to tools



import datetime
from collections.abc import Generator
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from pytest import MonkeyPatch

from appdog._internal.registry import Registry, RegistryPaths, _get_registry_paths
from appdog._internal.specs import AppSpec, EndpointInfo


class TestRegistryPaths:
    """Tests for RegistryPaths named tuple."""

    def test_init(self) -> None:
        """Test initialization of RegistryPaths."""
        paths = RegistryPaths(dir=Path('/test/dir'), specs=Path('/test/dir/registry.json'))
        assert paths.dir == Path('/test/dir')
        assert paths.specs == Path('/test/dir/registry.json')

    def test_immutable(self) -> None:
        """Test that RegistryPaths is immutable."""
        paths = RegistryPaths(dir=Path('/test/dir'), specs=Path('/test/dir/registry.json'))
        with pytest.raises(AttributeError):
            paths.dir = Path('/new/dir')  # type: ignore
        with pytest.raises(AttributeError):
            paths.specs = Path('/new/dir/registry.json')  # type: ignore


class TestRegistry:
    """Tests for Registry class."""

    @pytest.fixture
    def registry_dir(self, tmp_path: Path) -> Path:
        """Create a temporary registry directory."""
        registry_dir = tmp_path / 'registry'
        registry_dir.mkdir()
        return registry_dir

    @pytest.fixture
    def app_spec(self) -> AppSpec:
        """Create a test app specification."""
        return AppSpec(
            uri='http://example.com/api',
            data={'openapi': '3.0.0'},
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='test_hash',
        )

    @pytest.fixture
    def registry(self, registry_dir: Path) -> Generator[Registry, None, None]:
        """Create a Registry instance for testing."""
        from unittest.mock import patch

        # First create a real Registry instance
        registry = Registry(registry_dir)

        # Create a comprehensive mock for specs
        mock_specs = MagicMock()
        mock_specs.__getitem__.return_value = AppSpec(
            uri='http://example.com/api',
            data={'openapi': '3.0.0'},
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='test_hash',
        )
        mock_specs.__contains__.return_value = True

        # Replace specs with our mock
        registry.specs = mock_specs

        # Patch file existence checks
        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = True
            # Patch directory creation
            with patch.object(Path, 'mkdir'):
                # Patch file removal
                with patch.object(Path, 'unlink'):
                    yield registry

    def test_init(self, registry_dir: Path) -> None:
        """Test initialization of Registry."""
        registry = Registry(registry_dir)
        assert registry.paths.dir == registry_dir
        assert registry.paths.specs == registry_dir / 'registry.json'

    def test_init_default_dir(self) -> None:
        """Test initialization with default directory."""
        with patch('appdog._internal.registry.get_registry_dir') as mock_get_registry_dir:
            mock_get_registry_dir.return_value = Path('/test/source/lib')
            registry = Registry()
            assert registry.paths.dir == Path('/test/source/lib')
            assert registry.paths.specs == Path('/test/source/lib/registry.json')

    async def test_context_manager(self, registry: Registry) -> None:
        """Test async context manager."""
        async with registry as context:
            assert context is registry

    async def test_context_manager_save(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test that context manager saves on exit."""
        # Patch the path.exists method to return True after save
        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = True
            registry.specs['test_app'] = app_spec
            async with registry:
                pass
            assert mock_exists.return_value is True

    async def test_context_manager_error(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test that context manager doesn't save on error."""
        registry.specs['test_app'] = app_spec
        with pytest.raises(ValueError):
            async with registry:
                raise ValueError('Test error')
        # No need to check file existence as we're mocking everything

    def test_load(self, registry_dir: Path) -> None:
        """Test loading registry."""
        registry = Registry.load(registry_dir)
        assert registry.paths.dir == registry_dir
        assert registry.paths.specs == registry_dir / 'registry.json'

    def test_load_default(self, monkeypatch: MonkeyPatch) -> None:
        """Test loading default registry."""
        # Reset the singleton instance
        Registry._instance = None

        # First load should create a new instance
        registry1 = Registry.load()
        assert isinstance(registry1, Registry)

        # Second load should return the same instance
        registry2 = Registry.load()
        assert registry1 is registry2

        # Verify basic properties
        assert registry1.paths.dir is not None
        assert registry1.paths.specs is not None
        assert registry1.paths.specs.name == 'registry.json'

    def test_save(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test saving registry."""
        # Patch the path.exists method to return True after save
        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = True
            registry.specs['test_app'] = app_spec
            registry.save()
            assert mock_exists.return_value is True

    def test_add_app(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test adding an application."""
        # Patch the generator to avoid file operations
        with patch('appdog._internal.registry.generate_app_files') as mock_generate:
            # Make sure specs.__contains__ returns False for 'test_app' initially
            registry.specs.__contains__.side_effect = lambda key: False  # type: ignore

            # Call add_app
            registry.add_app('test_app', app_spec)

            # Verify generate_app_files was called
            mock_generate.assert_called_once()

            # Verify specs.__setitem__ was called with correct arguments
            registry.specs.__setitem__.assert_called_once_with('test_app', app_spec)  # type: ignore

    def test_add_app_existing(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test adding an existing application."""
        # Set up mocks to trigger the ValueError logic
        registry.specs.__contains__.return_value = True  # type: ignore
        registry.specs.__getitem__.return_value = app_spec  # type: ignore

        different_spec = AppSpec(
            uri='http://different.com/api',
            data={'openapi': '3.0.0'},
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='different_hash',
        )

        # The validation should fail when URIs don't match
        with pytest.raises(ValueError):
            registry.add_app('test_app', different_spec)

    def test_add_app_force(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test forcing addition of existing application."""
        # Patch the generator to avoid file operations
        with patch('appdog._internal.registry.generate_app_files') as mock_generate:
            # Set up mocks for existing app
            registry.specs.__contains__.return_value = True  # type: ignore
            registry.specs.__getitem__.return_value = app_spec  # type: ignore

            new_spec = AppSpec(
                uri='http://different.com/api',
                data={'openapi': '3.0.0'},
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                hash='new_hash',
            )

            # Mock the __setitem__ to update the return value of __getitem__
            def update_mock_return_value(key: str, value: AppSpec) -> None:
                registry.specs.__getitem__.return_value = value  # type: ignore

            registry.specs.__setitem__.side_effect = update_mock_return_value  # type: ignore

            # Force add should succeed despite URI mismatch
            registry.add_app('test_app', new_spec, force=True)

            # Verify specs.__setitem__ was called with correct arguments
            registry.specs.__setitem__.assert_called_with('test_app', new_spec)  # type: ignore

            # Verify registry.specs['test_app'] returns new_spec after the update
            assert registry.specs.__getitem__.return_value == new_spec  # type: ignore

            # Verify generate_app_files was called
            mock_generate.assert_called_once()

    def test_add_app_upgrade(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test upgrading an application."""
        # Patch the generator to avoid file operations
        with patch('appdog._internal.registry.generate_app_files') as mock_generate:
            # Set up mocks for existing app
            registry.specs.__contains__.return_value = True  # type: ignore
            registry.specs.__getitem__.return_value = app_spec  # type: ignore

            new_spec = AppSpec(
                uri=app_spec.uri,
                data=app_spec.data,
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                hash='new_hash',
            )

            # Mock the __setitem__ to update the return value of __getitem__
            def update_mock_return_value(key: str, value: AppSpec) -> None:
                registry.specs.__getitem__.return_value = value  # type: ignore

            registry.specs.__setitem__.side_effect = update_mock_return_value  # type: ignore

            # Upgrade should succeed with same URI but different hash
            registry.add_app('test_app', new_spec, upgrade=True)

            # Verify specs.__setitem__ was called with correct arguments
            registry.specs.__setitem__.assert_called_with('test_app', new_spec)  # type: ignore

            # Verify registry.specs['test_app'] returns new_spec after the update
            assert registry.specs.__getitem__.return_value == new_spec  # type: ignore

            # Verify generate_app_files was called
            mock_generate.assert_called_once()

    def test_remove_app(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test removing an application."""
        # Set up mocks for existing app
        registry.specs.__contains__.return_value = True  # type: ignore

        # Patch Path.exists specifically for the app directory check
        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = False

            # Remove the app
            registry.remove_app('test_app')

            # Verify __delitem__ was called
            registry.specs.__delitem__.assert_called_with('test_app')  # type: ignore

            # We're asserting the file doesn't exist after removal
            assert not mock_exists.return_value

    def test_remove_app_not_found(self, registry: Registry) -> None:
        """Test removing a non-existent application."""
        # Set up mocks for non-existent app
        registry.specs.__contains__.return_value = False  # type: ignore

        registry.remove_app('nonexistent_app')  # Should not raise an error

        # Verify no operations occurred
        assert not registry.specs.__delitem__.called  # type: ignore

    def test_lookup(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test looking up endpoints."""
        # Set up test endpoints for each app
        app1_endpoints = [
            EndpointInfo(
                name='get_users',
                method='get',
                path='/users',
                tags=['users'],
                operation_id='getUsers',
                summary='Get users',
                description='Get all users',
                parameters=[],
                request_body=None,
                responses={},
            )
        ]

        app2_endpoints = [
            EndpointInfo(
                name='get_items',
                method='get',
                path='/items',
                tags=['items'],
                operation_id='getItems',
                summary='Get items',
                description='Get all items',
                parameters=[],
                request_body=None,
                responses={},
            )
        ]

        # Mock the registry specs
        mock_specs = MagicMock()
        registry.specs = mock_specs

        # Set up a dictionary for the specs.items method
        mock_specs_dict = {'app1': MagicMock(), 'app2': MagicMock()}
        mock_specs.items.return_value = mock_specs_dict.items()

        # Set up the lookup method for each mock spec
        mock_specs_dict['app1'].lookup.return_value = app1_endpoints
        mock_specs_dict['app2'].lookup.return_value = app2_endpoints

        # Test lookup with no filters
        endpoints = registry.lookup()

        # Verify the correct endpoints were returned
        assert len(endpoints) == 2
        assert endpoints['app1'] == app1_endpoints
        assert endpoints['app2'] == app2_endpoints

        # Verify lookup was called with empty kwargs for each app
        assert mock_specs_dict['app1'].lookup.called
        assert mock_specs_dict['app2'].lookup.called

    def test_lookup_with_filters(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test looking up endpoints with filters."""
        # Create mock specs dictionary for the test
        mock_specs = {'app1': app_spec, 'app2': app_spec}

        # Set up test endpoints
        app1_endpoints = [
            EndpointInfo(
                name='get_users',
                method='get',
                path='/users',
                tags=['users'],
                operation_id='getUsers',
                summary='Get users',
                description='Get all users',
                parameters=[],
                request_body=None,
                responses={},
            )
        ]

        # Mock the registry specs dictionary and its methods
        registry.specs = MagicMock()
        registry.specs.items.return_value = mock_specs.items()

        # Mock the lookup method
        with patch(
            'appdog._internal.specs.AppSpec.lookup', return_value=app1_endpoints
        ) as mock_lookup:
            # Test lookup with app-specific filters
            endpoints = registry.lookup(app1=True)

            # Verify correct endpoints were returned
            assert len(endpoints) == 1
            assert 'app1' in endpoints
            assert 'app2' not in endpoints

            # Reset the mock for the next test
            mock_lookup.reset_mock()

            # Test lookup with a specific lookup config
            lookup_config = {'include_methods': ['get']}
            endpoints = registry.lookup(app1=lookup_config)  # type: ignore

            # Verify correct endpoints were returned
            assert len(endpoints) == 1
            assert 'app1' in endpoints
            assert 'app2' not in endpoints

            # Verify lookup was called with correct parameters
            mock_lookup.assert_called_once_with(**lookup_config)

    def test_lookup_with_kwargs(self, registry: Registry, app_spec: AppSpec) -> None:
        """Test looking up endpoints with kwargs filters."""
        # Setup mock for specs.items
        app_endpoints = [
            EndpointInfo(
                name='get_users',
                method='get',
                path='/users',
                tags=['users'],
                operation_id='getUsers',
                summary='Get users',
                description='',
                parameters=[],
                request_body=None,
                responses={},
            )
        ]

        # Mock the lookup method using patch instead of direct assignment (AppSpec is frozen)
        with patch.object(AppSpec, 'lookup', return_value=app_endpoints):
            # Set up the specs items mock
            registry.specs.items.return_value = [('test_app', app_spec)]  # type: ignore

            # Test lookup with path filter in filters
            endpoints = registry.lookup(test_app={'filters': {'path': r'/users'}})
            assert len(endpoints['test_app']) == 1


def test_get_registry_paths() -> None:
    """Test getting registry paths."""
    paths = _get_registry_paths(Path('/test/dir'))
    assert paths.dir == Path('/test/dir')
    assert paths.specs == Path('/test/dir/registry.json')


def test_get_registry_paths_default() -> None:
    """Test getting registry paths with default directory."""
    with patch('appdog._internal.registry.get_registry_dir') as mock_get_registry_dir:
        mock_get_registry_dir.return_value = Path('/test/source/lib')
        paths = _get_registry_paths()
        assert paths.dir == Path('/test/source/lib')
        assert paths.specs == Path('/test/source/lib/registry.json')



import datetime
import warnings
from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from pydantic import PydanticDeprecatedSince20

from appdog._internal.generator import (
    _generate_client_file,
    _generate_init_file,
    _generate_models_file,
    _get_path_params,
    _get_python_type,
    _get_query_params,
    _get_request_body_type,
    _get_response_model,
    _get_response_type,
    _has_query_params,
    generate_app_files,
)
from appdog._internal.specs import AppSpec, EndpointInfo


# Suppress Pydantic deprecation warnings from datamodel-code-generator
@pytest.fixture(autouse=True)
def ignore_pydantic_warnings() -> Generator[None, None, None]:
    """Suppress Pydantic deprecation warnings from the datamodel-code-generator library."""
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=PydanticDeprecatedSince20)
        yield


class TestGenerator:
    """Tests for generator functions."""

    @pytest.fixture
    def output_dir(self, tmp_path: Path) -> Path:
        """Create a temporary output directory."""
        return tmp_path

    @pytest.fixture
    def fixture_path(self) -> Path:
        """Fixture providing path to test fixtures."""
        return Path(__file__).parent / 'fixtures'

    @pytest.fixture
    def basic_spec_data(self, fixture_path: Path) -> dict:
        """Load basic test spec from fixture file."""
        with open(fixture_path / 'spec_basic.yaml') as f:
            return yaml.safe_load(f)  # type: ignore

    @pytest.fixture
    def petstore_spec_data(self, fixture_path: Path) -> dict:
        """Load petstore test spec from fixture file."""
        with open(fixture_path / 'spec_petstore.yaml') as f:
            return yaml.safe_load(f)  # type: ignore

    @pytest.fixture
    def basic_app_spec(self, basic_spec_data: dict) -> AppSpec:
        """Create a test app specification using basic fixture."""
        return AppSpec(
            uri='http://example.com/api',
            data=basic_spec_data,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='test_hash',
        )

    @pytest.fixture
    def petstore_app_spec(self, petstore_spec_data: dict) -> AppSpec:
        """Create a test app specification using petstore fixture."""
        return AppSpec(
            uri='http://example.com/api',
            data=petstore_spec_data,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='test_hash',
        )

    @pytest.fixture
    def endpoint_info(self) -> EndpointInfo:
        """Create a test endpoint info."""
        return EndpointInfo(
            name='get_test',
            method='get',
            path='/test',
            tags=[],
            operation_id='test',
            summary='Test endpoint',
            description='',
            parameters=[],
            request_body=None,
            responses={
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {
                            'schema': {
                                'type': 'object',
                                'properties': {
                                    'message': {'type': 'string', 'description': 'Test message'}
                                },
                            }
                        }
                    },
                }
            },
        )

    @pytest.fixture
    def param_string(self) -> dict:
        """Parameter with string type."""
        return {'name': 'test_param', 'schema': {'type': 'string'}}

    @pytest.fixture
    def param_integer(self) -> dict:
        """Parameter with integer type."""
        return {'name': 'test_param', 'schema': {'type': 'integer'}}

    @pytest.fixture
    def param_boolean(self) -> dict:
        """Parameter with boolean type."""
        return {'name': 'test_param', 'schema': {'type': 'boolean'}}

    @pytest.fixture
    def param_number(self) -> dict:
        """Parameter with number type."""
        return {'name': 'test_param', 'schema': {'type': 'number'}}

    @pytest.fixture
    def param_unknown(self) -> dict:
        """Parameter with unknown type."""
        return {'name': 'test_param', 'schema': {'type': 'unknown'}}

    @pytest.fixture
    def param_no_schema(self) -> dict:
        """Parameter without schema."""
        return {'name': 'test_param', 'type': 'string'}

    @pytest.fixture
    def endpoint_with_params(self) -> EndpointInfo:
        """Endpoint with various parameter types."""
        return EndpointInfo(
            name='test_params',
            method='get',
            path='/users/{user_id}/items/{item_id}',
            tags=[],
            operation_id='test',
            summary='Test params',
            description='',
            parameters=[
                {'name': 'user_id', 'in': 'path', 'required': True, 'schema': {'type': 'string'}},
                {'name': 'item_id', 'in': 'path', 'required': True, 'schema': {'type': 'integer'}},
                {'name': 'filter', 'in': 'query', 'required': False, 'schema': {'type': 'string'}},
                {'name': 'limit', 'in': 'query', 'required': False, 'schema': {'type': 'integer'}},
            ],
            request_body=None,
            responses={},
        )

    @pytest.fixture
    def endpoint_with_ref_response(self) -> EndpointInfo:
        """Endpoint with response using $ref."""
        return EndpointInfo(
            name='test_ref_response',
            method='get',
            path='/test',
            tags=[],
            operation_id='test',
            summary='Test ref response',
            description='',
            parameters=[],
            request_body=None,
            responses={
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {'schema': {'$ref': '#/components/schemas/TestModel'}}
                    },
                }
            },
        )

    @pytest.fixture
    def endpoint_with_object_response(self) -> EndpointInfo:
        """Endpoint with direct object response."""
        return EndpointInfo(
            name='test_object_response',
            method='get',
            path='/test',
            tags=[],
            operation_id='test',
            summary='Test object response',
            description='',
            parameters=[],
            request_body=None,
            responses={
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {
                            'schema': {
                                'type': 'object',
                                'properties': {'message': {'type': 'string'}},
                            }
                        }
                    },
                }
            },
        )

    @pytest.fixture
    def endpoint_with_request_body_ref(self) -> EndpointInfo:
        """Endpoint with request body using $ref."""
        return EndpointInfo(
            name='test_request_body_ref',
            method='post',
            path='/test',
            tags=[],
            operation_id='test',
            summary='Test request body ref',
            description='',
            parameters=[],
            request_body={
                'content': {
                    'application/json': {'schema': {'$ref': '#/components/schemas/TestModel'}}
                }
            },
            responses={},
        )

    @pytest.fixture
    def endpoint_with_request_body_direct(self) -> EndpointInfo:
        """Endpoint with direct object request body."""
        return EndpointInfo(
            name='test_request_body_direct',
            method='post',
            path='/test',
            tags=[],
            operation_id='test',
            summary='Test request body direct',
            description='',
            parameters=[],
            request_body={
                'content': {
                    'application/json': {
                        'schema': {'type': 'object', 'properties': {'message': {'type': 'string'}}}
                    }
                }
            },
            responses={},
        )

    def test_generate_client_file(self, output_dir: Path, endpoint_info: EndpointInfo) -> None:
        """Test generating client file."""
        # Setup
        name = 'test_app'
        endpoints = [endpoint_info]  # Now a list instead of dict

        # Create the app directory
        app_dir = output_dir / name
        app_dir.mkdir(parents=True, exist_ok=True)

        # Run test
        _generate_client_file(
            name,
            output_dir,
            uri='http://example.com/api',
            timestamp='2023-01-01T00:00:00Z',
            title='Test App',
            class_name='TestAppClient',
            base_url='http://example.com',
            endpoints=endpoints,
        )

        # Assert
        client_file = output_dir / name / 'client.py'
        assert client_file.exists(), 'Client file should exist'

        content = client_file.read_text()
        assert 'class TestAppClient(BaseClient)' in content
        assert 'async def get_test(' in content
        assert 'from appdog._internal.clients import BaseClient' in content
        assert 'return await self._get(path, **kwargs)' in content

    def test_generate_init_file(self, output_dir: Path) -> None:
        """Test generating init file."""
        # Setup
        name = 'test_app'

        # Create the app directory
        app_dir = output_dir / name
        app_dir.mkdir(parents=True, exist_ok=True)

        # Run test
        _generate_init_file(
            name,
            output_dir,
            uri='http://example.com/api',
            timestamp='2023-01-01T00:00:00Z',
            title='Test App',
            class_name='TestAppClient',
        )

        # Assert
        init_file = output_dir / name / '__init__.py'
        assert init_file.exists(), 'Init file should exist'

        content = init_file.read_text()
        assert 'from .client import TestAppClient' in content

    def test_generate_models_file(self, output_dir: Path, petstore_spec_data: dict) -> None:
        """Test generating models file using petstore fixture."""
        # Setup
        name = 'test_app_models'

        # Create the app directory
        app_dir = output_dir / name
        app_dir.mkdir(parents=True, exist_ok=True)

        # Run test
        _generate_models_file(
            name,
            output_dir,
            uri='http://example.com/api',
            timestamp='2023-01-01T00:00:00Z',
            title='Test App',
            data=petstore_spec_data,
        )

        # Assert
        models_file = output_dir / name / 'models.py'
        assert models_file.exists(), 'Models file should exist'

    def test_generate_models_file_error(self, output_dir: Path) -> None:
        """Test generating models file with error."""
        # Setup
        name = 'test_app_error'
        data = {'invalid': 'schema'}

        # Create the app directory
        app_dir = output_dir / name
        app_dir.mkdir(parents=True, exist_ok=True)

        # Run test
        _generate_models_file(
            name,
            output_dir,
            uri='http://example.com/api',
            timestamp='2023-01-01T00:00:00Z',
            title='Test App',
            data=data,
        )

        # Assert
        models_file = output_dir / name / 'models.py'
        assert models_file.exists(), 'Models file should exist'

        content = models_file.read_text()
        assert 'Error generating models' in content

    def test_generate_app_files(self, output_dir: Path, basic_app_spec: AppSpec) -> None:
        """Test generating app files using basic fixture."""
        # Run test
        app_name = 'test_app_gen1'
        generate_app_files(app_name, basic_app_spec, output_dir)

        # Assert
        app_dir = output_dir / app_name
        assert app_dir.exists(), 'App directory should exist'
        assert (app_dir / '__init__.py').exists(), 'Init file should exist'
        assert (app_dir / 'client.py').exists(), 'Client file should exist'
        assert (app_dir / 'models.py').exists(), 'Models file should exist'

    def test_generate_app_files_existing_dir(
        self, output_dir: Path, basic_app_spec: AppSpec
    ) -> None:
        """Test generating app files with existing directory."""
        # Create existing directory
        app_name = 'test_app_gen2'
        app_dir = output_dir / app_name
        app_dir.mkdir(exist_ok=True)
        (app_dir / 'existing_file.txt').write_text('Test')

        # Run test
        generate_app_files(app_name, basic_app_spec, output_dir, overwrite=True)

        # Assert
        assert app_dir.exists(), 'App directory should exist'
        assert not (app_dir / 'existing_file.txt').exists(), 'Existing file should be removed'

    def test_generate_app_files_existing_dir_no_overwrite(
        self, output_dir: Path, basic_app_spec: AppSpec
    ) -> None:
        """Test generating app files with existing directory without overwrite."""
        # Create existing directory
        app_name = 'test_app_gen3'
        app_dir = output_dir / app_name
        app_dir.mkdir(exist_ok=True)

        # Run test and assert error
        with pytest.raises(ValueError, match='Output directory already exists'):
            generate_app_files(app_name, basic_app_spec, output_dir, overwrite=False)

    def test_generate_app_files_error(self, output_dir: Path, basic_app_spec: AppSpec) -> None:
        """Test generating app files with error."""
        app_name = 'test_app_gen4'

        # Mock _generate_client_file to raise an error
        with patch(
            'appdog._internal.generator._generate_client_file',
            side_effect=ValueError('Test error'),
        ):
            # Run test and assert error
            with pytest.raises(ValueError, match='Test error'):
                generate_app_files(app_name, basic_app_spec, output_dir)

            # Assert directory was cleaned up
            app_dir = output_dir / app_name
            assert not app_dir.exists(), 'App directory should be removed on error'

    def test_get_python_type(
        self,
        param_string: dict,
        param_integer: dict,
        param_boolean: dict,
        param_number: dict,
        param_unknown: dict,
        param_no_schema: dict,
    ) -> None:
        """Test _get_python_type function."""
        assert _get_python_type(param_string) == 'str'
        assert _get_python_type(param_integer) == 'int'
        assert _get_python_type(param_boolean) == 'bool'
        assert _get_python_type(param_number) == 'float'
        assert _get_python_type(param_unknown) == 'Any'
        assert _get_python_type(param_no_schema) == 'str'
        assert _get_python_type({}) == 'Any'  # Empty param

    def test_get_response_model(
        self,
        endpoint_info: EndpointInfo,
        endpoint_with_ref_response: EndpointInfo,
        endpoint_with_object_response: EndpointInfo,
    ) -> None:
        """Test _get_response_model function."""
        # Basic endpoint with inline schema - should return None
        assert _get_response_model(endpoint_info) is None

        # Endpoint with $ref response - should return model name with 'models.' prefix
        assert _get_response_model(endpoint_with_ref_response) == 'models.TestModel'

        # Endpoint with direct object response - should return None
        assert _get_response_model(endpoint_with_object_response) is None

    def test_get_response_type(
        self,
        endpoint_info: EndpointInfo,
        endpoint_with_ref_response: EndpointInfo,
        endpoint_with_object_response: EndpointInfo,
    ) -> None:
        """Test _get_response_type function."""
        # Basic endpoint with inline schema - should return dict type
        assert _get_response_type(endpoint_info) == 'dict[str, Any]'

        # Endpoint with $ref response - should return model type
        assert _get_response_type(endpoint_with_ref_response) == 'models.TestModel'

        # Endpoint with direct object response - should return dict type
        assert _get_response_type(endpoint_with_object_response) == 'dict[str, Any]'

    def test_has_query_params(
        self, endpoint_info: EndpointInfo, endpoint_with_params: EndpointInfo
    ) -> None:
        """Test _has_query_params function."""
        # Basic endpoint with no query params
        assert _has_query_params(endpoint_info) is False

        # Endpoint with query params
        assert _has_query_params(endpoint_with_params) is True

    def test_get_path_params(
        self, endpoint_info: EndpointInfo, endpoint_with_params: EndpointInfo
    ) -> None:
        """Test _get_path_params function."""
        # Basic endpoint with no path params
        assert len(_get_path_params(endpoint_info)) == 0

        # Endpoint with path params
        path_params = _get_path_params(endpoint_with_params)
        assert len(path_params) == 2
        assert path_params[0]['name'] == 'user_id'
        assert path_params[1]['name'] == 'item_id'

    def test_get_query_params(
        self, endpoint_info: EndpointInfo, endpoint_with_params: EndpointInfo
    ) -> None:
        """Test _get_query_params function."""
        # Basic endpoint with no query params
        assert len(_get_query_params(endpoint_info)) == 0

        # Endpoint with query params
        query_params = _get_query_params(endpoint_with_params)
        assert len(query_params) == 2
        assert query_params[0]['name'] == 'filter'
        assert query_params[1]['name'] == 'limit'

    def test_get_request_body_type(
        self,
        endpoint_with_request_body_ref: EndpointInfo,
        endpoint_with_request_body_direct: EndpointInfo,
    ) -> None:
        """Test _get_request_body_type function."""
        # Endpoint with $ref request body
        assert _get_request_body_type(endpoint_with_request_body_ref) == 'models.TestModel'

        # Endpoint with direct object request body
        assert _get_request_body_type(endpoint_with_request_body_direct) == 'dict[str, Any]'



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from pathlib import Path

from .project import Project
from .registry import Registry


@asynccontextmanager
async def project_manager(
    project_dir: Path | str | None = None,
    registry_dir: Path | str | None = None,
) -> AsyncGenerator[Project, None]:
    """Context manager for project that automatically saves when exiting.

    Args:
        project_dir: Path to project directory (default: auto-detect)
        registry_dir: Path to registry directory (default: auto-detect)
    """
    async with Project.load(project_dir, registry_dir) as context:
        yield context


@asynccontextmanager
async def registry_manager(
    registry_dir: Path | str | None = None,
) -> AsyncGenerator[Registry, None]:
    """Context manager for registry that automatically saves when exiting.

    Args:
        registry_dir: Path to registry directory (default: auto-detect)
    """
    async with Registry.load(registry_dir) as context:
        yield context



import json
import typing
from pathlib import Path
from typing import Any, TypeVar

import httpx
from httpx import AsyncClient, Timeout
from typing_extensions import Self, TypedDict, Unpack

from .errors import (
    AuthError,
    NotFoundError,
    RateLimitError,
    RequestError,
    ResponseError,
)
from .project import Project
from .settings import ClientSettings

_T = TypeVar('_T')


class ClientConfig(TypedDict, total=False):
    """Client arguments."""

    base_url: str | None
    """Base URL for the client."""

    api_key: str | None
    """API key for authentication."""

    token: str | None
    """Bearer token for authentication."""

    api_key_header: str | None
    """Header name for API key."""

    token_header: str | None
    """Header name for bearer token."""

    timeout: float | None
    """Request timeout in seconds."""


class BaseClient:
    """Base client for API services."""

    paths: dict[str, str] = {}
    """Paths for the client."""

    def __init__(self, name: str, **config: Unpack[ClientConfig]) -> None:
        base_url = config.pop('base_url', None)
        if not base_url:
            raise ValueError('A base URL is required')
        self._client: AsyncClient | None = None
        self._settings = ClientSettings.with_env_prefix(name=name, base_url=base_url, **config)

    async def __aenter__(self) -> Self:
        self.open()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    @classmethod
    def load(cls, name: str, project_dir: Path | str | None = None) -> Self:
        """Load the client from the project directory."""
        project = Project.load(project_dir)
        if name not in project.settings:
            return cls(name)
        config = project.settings[name].get_client_config()
        return cls(name, **config)

    def open(self) -> None:
        """Open the HTTP client."""
        if self._client is None:
            self._client = self._create_client()

    async def close(self) -> None:
        """Close the HTTP client."""
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    def _create_client(self) -> AsyncClient:
        """Create a new HTTP client."""
        return AsyncClient(
            base_url=self._settings.base_url,
            headers=self._settings.get_headers(),
            timeout=self._settings.timeout,
        )

    async def _request(  # noqa: C901
        self,
        method: str,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        data: Any | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make an API request.

        Args:
            method: HTTP method
            path: URL path
            params: Query parameters
            data: Request body data
            headers: Additional headers
            return_type: Optional Pydantic model to parse the response

        Returns:
            API response data, optionally parsed into the provided model

        Raises:
            ClientError: If the request fails.
        """
        client = self._client
        close_on_exit = False
        if client is None:
            client = self._create_client()
            close_on_exit = True

        timeout = Timeout(self._settings.timeout, connect=5.0)

        if not path.startswith('/'):
            path = f'/{path}'
        headers = headers or {}
        headers = {**self._settings.get_headers(), **headers}
        if data is not None and hasattr(data, 'model_dump') and callable(data.model_dump):
            data = data.model_dump(mode='json')

        try:
            response = await client.request(
                method=method,
                url=path,
                params=params,
                json=data,
                headers=headers,
                timeout=timeout,
            )

            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                error_detail = str(e)
                try:
                    error_json = response.json()
                    if isinstance(error_json, dict) and 'message' in error_json:
                        error_detail = error_json['message']
                except json.JSONDecodeError:
                    pass

                if response.status_code in (401, 403):
                    raise AuthError(
                        f'Authentication error: {error_detail}',
                        status_code=response.status_code,
                        response=response,
                    ) from e
                elif response.status_code == 404:
                    raise NotFoundError(
                        f'Resource not found: {error_detail}',
                        status_code=response.status_code,
                        response=response,
                    ) from e
                elif response.status_code == 429:
                    raise RateLimitError(
                        f'Rate limit exceeded: {error_detail}',
                        status_code=response.status_code,
                        response=response,
                    ) from e
                elif 400 <= response.status_code < 500:
                    raise ResponseError(
                        f'Client error: {error_detail}',
                        status_code=response.status_code,
                        response=response,
                    ) from e
                else:
                    raise ResponseError(
                        f'Server error: {error_detail}',
                        status_code=response.status_code,
                        response=response,
                    ) from e

            try:
                response_data = response.json()
            except json.JSONDecodeError:
                response_data = response.text

            if return_type is None:
                return response_data

            try:
                constructor = typing.cast(Any, return_type)
                if hasattr(constructor, 'model_validate') and callable(constructor.model_validate):
                    response_data = constructor.model_validate(response_data)
                elif isinstance(response_data, dict):
                    response_data = constructor(**response_data)
                else:
                    response_data = constructor(response_data)
            except Exception as e:
                raise ResponseError(
                    f'Validation error: {e}',
                    status_code=response.status_code,
                    response=response,
                ) from e

            return response_data

        except httpx.RequestError as e:
            raise RequestError('Request failed') from e
        finally:
            if close_on_exit:
                await client.aclose()

    async def _get(
        self,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make a GET request."""
        return await self._request(
            'GET',
            path,
            params=params,
            headers=headers,
            return_type=return_type,
        )

    async def _post(
        self,
        path: str,
        *,
        data: Any | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make a POST request."""
        return await self._request(
            'POST',
            path,
            params=params,
            data=data,
            headers=headers,
            return_type=return_type,
        )

    async def _put(
        self,
        path: str,
        *,
        data: Any | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make a PUT request."""
        return await self._request(
            'PUT',
            path,
            params=params,
            data=data,
            headers=headers,
            return_type=return_type,
        )

    async def _patch(
        self,
        path: str,
        *,
        data: Any | None = None,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make a PATCH request."""
        return await self._request(
            'PATCH',
            path,
            params=params,
            data=data,
            headers=headers,
            return_type=return_type,
        )

    async def _delete(
        self,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        return_type: type[_T] | None = None,
    ) -> Any | _T:
        """Make a DELETE request."""
        return await self._request(
            'DELETE',
            path,
            params=params,
            headers=headers,
            return_type=return_type,
        )



import re

CHAR_MARK = '\ue000'
"""A special character used as a placeholder for string transformations."""

CHAR_SEP = '\ue001'
"""A special character used as a separator for string transformations."""


def to_pascal_case(string: str) -> str:
    """Convert a string to pascal case."""
    if not string:
        return string
    s = string
    s = re.sub(r'[\s\-\_]', CHAR_SEP, s)
    s = re.sub(r'(?<![^a-zA-Z\d])(?=[A-Z][a-z\d]+)', CHAR_SEP, s)
    s = re.sub(rf'^{CHAR_SEP}+|{CHAR_SEP}+$', '', s)
    s = re.sub(rf'{CHAR_SEP}+(.)', lambda m: m.group(1).upper(), s)
    s = re.sub(r'(?:^)(?=[a-zA-Z\d]+)', CHAR_MARK, s)
    s = re.sub(r'(?<=[^a-zA-Z\d])(?=[a-zA-Z\d]+)', CHAR_MARK, s)
    s = re.sub(rf'{CHAR_MARK}+(.)', lambda m: m.group(1).upper(), s)
    return s


def to_snake_case(string: str) -> str:
    """Convert a string to snake case."""
    if not string:
        return string
    s = string
    s = re.sub(r'[\s\-\_]', CHAR_SEP, s)
    s = re.sub(r'(?<![^a-zA-Z\d])(?=[A-Z][a-z\d]+)', CHAR_SEP, s)
    s = re.sub(rf'^{CHAR_SEP}+|{CHAR_SEP}+$', '', s)
    s = re.sub(rf'{CHAR_SEP}+', '_', s)
    return s.lower()


def to_title_case(string: str) -> str:
    """Convert a string to title case."""
    if not string:
        return string
    s = string
    s = re.sub(r'[\s\-\_]', CHAR_SEP, s)
    s = re.sub(r'(?<![^a-zA-Z\d])(?=[A-Z][a-z\d]+)', CHAR_SEP, s)
    s = re.sub(rf'^{CHAR_SEP}+|{CHAR_SEP}+$', '', s)
    s = re.sub(rf'{CHAR_SEP}+(.)', lambda m: ' ' + m.group(1).upper(), s)
    s = re.sub(r'(?:^)(?=[a-zA-Z\d]+)', CHAR_MARK, s)
    s = re.sub(r'(?<=[^a-zA-Z\d])(?=[a-zA-Z\d]+)', CHAR_MARK, s)
    s = re.sub(rf'{CHAR_MARK}+(.)', lambda m: m.group(1).upper(), s)
    return s


def to_name_case(string: str) -> str:
    """Convert a string to name case."""
    if not string:
        return string
    pattern_brackets = r'\{[^\{]*?\}'
    pattern_path = r'[\.\/\\]'
    # Collect all parameters
    params = []
    for match in re.finditer(pattern_brackets, string):
        content = to_snake_case(match.group(0)[1:-1])
        params.append(content)
    # Convert the string
    s = string
    s = re.sub(pattern_brackets, CHAR_SEP, s)
    s = re.sub(pattern_path, CHAR_SEP, s)
    s = to_snake_case(s)
    for count, param in enumerate(params):
        s += '_by_' if count == 0 else '_and_'
        s += param
    return s



from typing import Any

import pytest


@pytest.fixture
def sample_data() -> dict[str, Any]:
    """Fixture providing sample data dictionaries for testing."""
    return {
        'simple': {'key': 'value'},
        'nested': {'outer': {'inner': 'value'}, 'list': [1, 2, 3]},
        'ordered': {'a': 1, 'b': 2, 'c': 3},
        'unordered': {'c': 3, 'a': 1, 'b': 2},
    }


class AsyncNoWarningMock:
    """
    A mock for async functions that doesn't produce warnings.

    This is a simple callable that returns a value immediately instead of a coroutine,
    preventing the "coroutine was never awaited" warnings.
    """

    def __init__(self, return_value: Any = None) -> None:
        self.return_value = return_value
        self.call_count = 0
        self.call_args: Any | None = None
        self.call_args_list: list[Any] = []

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Record the call and return the configured return value."""
        self.call_count += 1
        self.call_args = (args, kwargs)
        self.call_args_list.append((args, kwargs))
        return self

    def __await__(self) -> Any:
        """Make this mock awaitable."""

        async def _awaitable() -> Any:
            return self.return_value

        return _awaitable().__await__()

    def assert_called_once(self) -> None:
        """Assert the mock was called exactly once."""
        if self.call_count != 1:
            raise AssertionError(f'Expected 1 call, got {self.call_count}')



import datetime
import json
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest
import yaml
from httpx import Response

from appdog._internal.specs import AppSpec, EndpointInfo


class TestEndpointInfo:
    """Tests for the EndpointInfo dataclass."""

    def test_init(self) -> None:
        """Test initialization of EndpointInfo."""
        endpoint = EndpointInfo(
            name='get_users',
            method='get',
            path='/users',
            tags=['users'],
            operation_id='getUsers',
            summary='Get users',
            description='Get all users',
            parameters=[{'name': 'limit', 'in': 'query', 'schema': {'type': 'integer'}}],
            request_body=None,
            responses={'200': {'description': 'OK'}},
        )

        assert endpoint.name == 'get_users'
        assert endpoint.method == 'get'
        assert endpoint.path == '/users'
        assert endpoint.tags == ['users']
        assert endpoint.operation_id == 'getUsers'
        assert endpoint.summary == 'Get users'
        assert endpoint.description == 'Get all users'
        assert endpoint.parameters == [
            {'name': 'limit', 'in': 'query', 'schema': {'type': 'integer'}}
        ]
        assert endpoint.request_body is None
        assert endpoint.responses == {'200': {'description': 'OK'}}


class TestAppSpec:
    """Tests for the AppSpec class."""

    @pytest.fixture
    def sample_data(self) -> dict[str, Any]:
        """Fixture providing sample OpenAPI data."""
        return {
            'openapi': '3.0.0',
            'info': {'title': 'Test API', 'version': '1.0.0'},
            'paths': {
                '/test': {
                    'get': {
                        'operationId': 'getTest',
                        'summary': 'Test endpoint',
                        'responses': {'200': {'description': 'Success'}},
                    }
                }
            },
        }

    @pytest.fixture
    def sample_spec(self, sample_data: dict[str, Any]) -> AppSpec:
        """Fixture providing a sample AppSpec."""
        return AppSpec(
            uri='http://example.com/openapi.json',
            data=sample_data,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='sample_hash',
        )

    @pytest.fixture
    def fixture_path(self) -> Path:
        """Fixture providing path to test fixtures."""
        return Path(__file__).parent / 'fixtures'

    def test_init(self, sample_data: dict[str, Any]) -> None:
        """Test initialization of AppSpec."""
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        spec = AppSpec(
            uri='http://example.com/openapi.json',
            data=sample_data,
            timestamp=timestamp,
            hash='sample_hash',
        )

        assert spec.uri == 'http://example.com/openapi.json'
        assert spec.data == sample_data
        assert spec.timestamp == timestamp
        assert spec.hash == 'sample_hash'

    @pytest.mark.asyncio
    async def test_fetch_local_json(self, fixture_path: Path, tmp_path: Path) -> None:
        """Test fetching from a local JSON file."""
        # Create a sample JSON file
        json_file = tmp_path / 'spec.json'
        with open(json_file, 'w') as f:
            json.dump({'openapi': '3.0.0', 'info': {'title': 'Test API', 'version': '1.0.0'}}, f)

        # Fetch the specification
        spec = await AppSpec.fetch(str(json_file))

        assert spec.uri == str(json_file)
        assert spec.data == {'openapi': '3.0.0', 'info': {'title': 'Test API', 'version': '1.0.0'}}
        assert isinstance(spec.timestamp, datetime.datetime)
        assert isinstance(spec.hash, str)

    @pytest.mark.asyncio
    async def test_fetch_local_yaml(self, fixture_path: Path) -> None:
        """Test fetching from a local YAML file."""
        yaml_file = fixture_path / 'spec_basic.yaml'

        # Fetch the specification
        with patch('appdog._internal.specs.json.load') as mock_json_load:
            # Mock json.load to load YAML file since we're testing with YAML but code expects JSON
            with open(yaml_file) as f:
                yaml_content = yaml.safe_load(f)
            mock_json_load.return_value = yaml_content

            spec = await AppSpec.fetch(str(yaml_file))

        assert spec.uri == str(yaml_file)
        assert spec.data['openapi'] == '3.0.0'
        assert spec.data['info']['title'] == 'Test API'
        assert isinstance(spec.timestamp, datetime.datetime)
        assert isinstance(spec.hash, str)

    @pytest.mark.asyncio
    async def test_fetch_local_invalid(self, tmp_path: Path) -> None:
        """Test fetching from an invalid local file."""
        # Create an invalid JSON file
        invalid_file = tmp_path / 'invalid.json'
        with open(invalid_file, 'w') as f:
            f.write('invalid json')

        # Fetch the specification with error
        with pytest.raises(ValueError, match='Failed to parse local specification'):
            await AppSpec.fetch(str(invalid_file))

    @pytest.mark.asyncio
    async def test_fetch_http(self) -> None:
        """Test fetching from an HTTP URL."""
        sample_data = {
            'openapi': '3.0.0',
            'info': {'title': 'Test API', 'version': '1.0.0'},
        }

        # Mock the HTTP response
        mock_response = MagicMock(spec=Response)
        mock_response.json.return_value = sample_data
        mock_response.raise_for_status = MagicMock()

        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.return_value = mock_response

        with patch('httpx.AsyncClient', return_value=mock_client):
            spec = await AppSpec.fetch('http://example.com/openapi.json')

        assert spec.uri == 'http://example.com/openapi.json'
        assert spec.data == sample_data
        assert isinstance(spec.timestamp, datetime.datetime)
        assert isinstance(spec.hash, str)

    @pytest.mark.asyncio
    async def test_fetch_http_error(self) -> None:
        """Test fetching from an HTTP URL with error."""
        # Mock the HTTP error
        mock_client = AsyncMock()
        mock_client.__aenter__.return_value = mock_client
        mock_client.get.side_effect = httpx.HTTPError('HTTP Error')

        with patch('httpx.AsyncClient', return_value=mock_client):
            with pytest.raises(ValueError, match='Failed to load specification'):
                await AppSpec.fetch('http://example.com/openapi.json')

    @pytest.mark.asyncio
    async def test_fetch_invalid_uri(self) -> None:
        """Test fetching from an invalid URI."""
        with pytest.raises(ValueError, match='Failed to load specification from'):
            await AppSpec.fetch('https://example.com/openapi.json')

    def test_lookup_no_filters(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints without filters."""
        # Add a path to the spec
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                    'summary': 'Get users',
                    'parameters': [{'name': 'limit', 'in': 'query'}],
                    'responses': {'200': {'description': 'OK'}},
                }
            }
        }

        endpoints = sample_spec.lookup()

        assert len(endpoints) == 1
        endpoint = endpoints[0]
        assert endpoint.name == 'get_users'
        assert endpoint.method == 'get'
        assert endpoint.path == '/users'
        assert endpoint.tags == ['users']
        assert endpoint.operation_id == 'getUsers'
        assert endpoint.summary == 'Get users'

    def test_lookup_with_pattern(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with pattern filters."""
        # Add multiple paths to the spec
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                    'summary': 'Get users',
                }
            },
            '/pets': {
                'get': {
                    'operationId': 'getPets',
                    'tags': ['pets'],
                    'summary': 'Get pets',
                }
            },
        }

        # Look up with path pattern
        endpoints = sample_spec.lookup(filters={'path': r'/users'})

        assert len(endpoints) == 1
        assert endpoints[0].path == '/users'
        assert endpoints[0].name == 'get_users'

        # Test with no matches
        endpoints = sample_spec.lookup(filters={'path': r'/nonexistent'})
        assert len(endpoints) == 0

    def test_lookup_with_method_filters(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with method filters."""
        # Add multiple methods to the spec
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                    'summary': 'Get users',
                },
                'post': {
                    'operationId': 'createUser',
                    'tags': ['users'],
                    'summary': 'Create user',
                },
            }
        }

        # Test include_methods
        endpoints = sample_spec.lookup(include_methods=['get'])
        assert len(endpoints) == 1
        assert endpoints[0].method == 'get'
        assert endpoints[0].name == 'get_users'

        # Test exclude_methods
        endpoints = sample_spec.lookup(exclude_methods=['get'])
        assert len(endpoints) == 1
        assert endpoints[0].method == 'post'
        assert endpoints[0].name == 'post_users'

        # Test with LookupConfig
        lookup_config = {'include_methods': ['get']}
        endpoints = sample_spec.lookup(**lookup_config)  # type: ignore
        assert len(endpoints) == 1
        assert endpoints[0].method == 'get'
        assert endpoints[0].name == 'get_users'

    def test_lookup_with_tag_filters(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with tag filters."""
        # Add multiple tags to the spec
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users', 'public'],
                },
                'post': {
                    'operationId': 'createUser',
                    'tags': ['users', 'admin'],
                },
            }
        }

        # Look up with tag include filter
        endpoints = sample_spec.lookup(include_tags=['public'])
        assert len(endpoints) == 1
        assert 'public' in endpoints[0].tags
        assert endpoints[0].name == 'get_users'

        # Look up with tag exclude filter
        endpoints = sample_spec.lookup(exclude_tags=['admin'])
        assert len(endpoints) == 1
        assert 'admin' not in endpoints[0].tags
        assert endpoints[0].name == 'get_users'

    def test_lookup_with_petstore_spec(self, fixture_path: Path) -> None:
        """Test looking up endpoints with petstore spec."""
        # Load petstore spec
        import yaml

        with open(fixture_path / 'spec_petstore.yaml') as f:
            data = yaml.safe_load(f)

        # Create spec
        spec = AppSpec(
            uri='http://petstore.example.com',
            data=data,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            hash='petstore_hash',
        )

        # Look up endpoints
        endpoints = spec.lookup()

        # Verify endpoints
        assert len(endpoints) == 3
        endpoint_names = [e.name for e in endpoints]
        assert 'get_pets' in endpoint_names
        assert 'post_pets' in endpoint_names
        assert 'get_pets_by_pet_id' in endpoint_names

    def test_lookup_skip_non_operations(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints skips non-HTTP methods."""
        # Add parameters and other fields to path
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                },
                'parameters': [{'name': 'limit', 'in': 'query'}],
            }
        }

        endpoints = sample_spec.lookup()
        assert len(endpoints) == 1
        assert endpoints[0].name == 'get_users'

    def test_lookup_skip_empty_operations(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints skips empty operations."""
        # Add empty operation
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                },
                'post': None,
            }
        }

        endpoints = sample_spec.lookup()
        assert len(endpoints) == 1
        assert endpoints[0].name == 'get_users'

    def test_lookup_duplicate_endpoint(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with duplicate operation IDs."""
        # Add multiple operations with same ID
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                },
            },
            '/users2': {
                'get': {
                    'operationId': 'getUsers',
                },
            },
        }

        # This should now assert to check we have both endpoints despite the duplicate IDs
        endpoints = sample_spec.lookup()
        assert len(endpoints) == 2
        names = [e.name for e in endpoints]
        assert 'get_users' in names
        assert 'get_users2' in names

    def test_lookup_with_path_parameters(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with path parameters."""
        # Add paths with parameters
        sample_spec.data['paths'] = {
            '/users/{userId}': {
                'get': {
                    'operationId': 'getUserById',
                    'parameters': [{'name': 'userId', 'in': 'path', 'required': True}],
                },
            },
            '/users/{userId}/posts/{postId}': {
                'get': {
                    'operationId': 'getUserPost',
                    'parameters': [
                        {'name': 'userId', 'in': 'path', 'required': True},
                        {'name': 'postId', 'in': 'path', 'required': True},
                    ],
                },
            },
        }

        endpoints = sample_spec.lookup()
        assert len(endpoints) == 2
        names = [e.name for e in endpoints]
        assert 'get_users_by_user_id' in names
        assert 'get_users_posts_by_user_id_and_post_id' in names

    def test_lookup_with_kwargs_regex(self, sample_spec: AppSpec) -> None:
        """Test looking up endpoints with regex pattern in kwargs."""
        # Add multiple paths to the spec
        sample_spec.data['paths'] = {
            '/users': {
                'get': {
                    'operationId': 'getUsers',
                    'tags': ['users'],
                    'summary': 'Get all users',
                }
            },
            '/users/search': {
                'get': {
                    'operationId': 'searchUsers',
                    'tags': ['users'],
                    'summary': 'Search users',
                }
            },
            '/pets': {
                'get': {
                    'operationId': 'getPets',
                    'tags': ['pets'],
                    'summary': 'Get all pets',
                }
            },
        }

        # Look up with regex string in kwargs for path
        endpoints = sample_spec.lookup(filters={'path': r'/users.*'})
        assert len(endpoints) == 2
        paths = [e.path for e in endpoints]
        assert '/users' in paths
        assert '/users/search' in paths

        # Look up with regex string in kwargs for summary
        endpoints = sample_spec.lookup(filters={'summary': r'.*all.*'})
        assert len(endpoints) == 2
        summaries = [e.summary for e in endpoints]
        assert 'Get all users' in summaries
        assert 'Get all pets' in summaries

        # Look up with multiple filters
        endpoints = sample_spec.lookup(filters={'path': r'/users.*', 'summary': r'Search.*'})
        assert len(endpoints) == 1
        assert endpoints[0].path == '/users/search'

        # Using compiled regex pattern
        import re

        path_pattern = re.compile(r'/users.*')
        endpoints = sample_spec.lookup(filters={'path': path_pattern})
        assert len(endpoints) == 2

    def test_equality(self, sample_data: dict[str, Any]) -> None:
        """Test equality comparison between AppSpec instances."""
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        spec1 = AppSpec(
            uri='http://example.com/openapi.json',
            data=sample_data,
            timestamp=timestamp,
            hash='hash1',
        )

        # Same URI and hash
        spec2 = AppSpec(
            uri='http://example.com/openapi.json',
            data=sample_data,
            timestamp=datetime.datetime.now(datetime.timezone.utc),  # Different timestamp
            hash='hash1',
        )

        # Same URI, different hash
        spec3 = AppSpec(
            uri='http://example.com/openapi.json',
            data=sample_data,
            timestamp=timestamp,
            hash='hash2',
        )

        # Different URI
        spec4 = AppSpec(
            uri='http://example.com/another.json',
            data=sample_data,
            timestamp=timestamp,
            hash='hash1',
        )

        # Test equality
        assert spec1 == spec2  # Same URI and hash
        assert spec1 != spec3  # Same URI, different hash
        assert spec1 != spec4  # Different URI
        assert spec1 != 'not_a_spec'  # Different type

        # Test inequality
        assert not (spec1 != spec2)  # Same URI and hash



__version__ = '0.1.0-b3'

from ._internal.managers import project_manager, registry_manager
from ._internal.mcp import (
    MCPResolver,
    MCPStrategy,
    add_to_fastmcp,
    create_resource_info,
    create_tool_info,
    mount_to_fastmcp,
)
from ._internal.project import Project
from ._internal.registry import Registry

__all__ = (
    'MCPResolver',
    'MCPStrategy',
    'Project',
    'Registry',
    'add_to_fastmcp',
    'create_resource_info',
    'create_tool_info',
    'mount_to_fastmcp',
    'project_manager',
    'registry_manager',
)



from httpx import Response


class ClientError(Exception):
    def __init__(
        self,
        message: str,
        *,
        status_code: int | None = None,
        response: Response | None = None,
    ):
        self.status_code = status_code
        self.response = response
        super().__init__(message)


class AuthError(ClientError):
    pass


class NotFoundError(ClientError):
    pass


class RateLimitError(ClientError):
    pass


class RequestError(ClientError):
    pass


class ResponseError(ClientError):
    pass



import datetime
import hashlib
import json
import os
from pathlib import Path
from typing import Any


def compute_hash(data: dict[str, Any]) -> str:
    """Compute the hash of the given dictionary."""
    data_json = json.dumps(data, sort_keys=True)
    return hashlib.sha256(data_json.encode()).hexdigest()


def get_project_dir() -> Path:
    """Get the path to the project directory."""
    project_dir = os.environ.get('APPDOG_PROJECT')
    if project_dir:
        return Path(project_dir)
    return Path.cwd()


def get_registry_dir() -> Path:
    """Get the path to the registry directory."""
    registry_dir = os.environ.get('APPDOG_REGISTRY')
    if registry_dir:
        return Path(registry_dir)
    return get_source_dir()


def get_source_dir() -> Path:
    """Get the path to the applications source directory."""
    return Path(__file__).parent.parent


def get_timestamp() -> str:
    """Get the current timestamp."""
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')



import datetime
import hashlib
import json
import re
from pathlib import Path
from typing import Any

from appdog._internal.utils import compute_hash, get_registry_dir, get_timestamp


class TestComputeHash:
    def test_compute_hash_with_simple_dict(self, sample_data: dict[str, Any]) -> None:
        """Test compute_hash with a simple dictionary."""
        test_dict = sample_data['simple']
        expected = hashlib.sha256(json.dumps(test_dict, sort_keys=True).encode()).hexdigest()
        assert compute_hash(test_dict) == expected

    def test_compute_hash_with_nested_dict(self, sample_data: dict[str, Any]) -> None:
        """Test compute_hash with a nested dictionary."""
        test_dict = sample_data['nested']
        expected = hashlib.sha256(json.dumps(test_dict, sort_keys=True).encode()).hexdigest()
        assert compute_hash(test_dict) == expected

    def test_compute_hash_deterministic(self, sample_data: dict[str, Any]) -> None:
        """Test that compute_hash produces the same hash for the same data regardless of order."""
        ordered_dict = sample_data['ordered']
        unordered_dict = sample_data['unordered']
        assert compute_hash(ordered_dict) == compute_hash(unordered_dict)

    def test_compute_hash_different_for_different_data(self) -> None:
        """Test that compute_hash produces different hashes for different data."""
        dict1 = {'key': 'value1'}
        dict2 = {'key': 'value2'}
        assert compute_hash(dict1) != compute_hash(dict2)


class TestGetSourceDir:
    def test_get_registry_dir_returns_path(self) -> None:
        """Test that get_registry_dir returns a Path object."""
        registry_dir = get_registry_dir()
        assert isinstance(registry_dir, Path)

    def test_get_registry_dir_points_to_src(self) -> None:
        """Test that get_registry_dir points to the src directory."""
        registry_dir = get_registry_dir()
        assert registry_dir.name == 'appdog'
        assert registry_dir.is_dir()


class TestGetTimestamp:
    def test_get_timestamp_returns_string(self) -> None:
        """Test that get_timestamp returns a string."""
        timestamp = get_timestamp()
        assert isinstance(timestamp, str)

    def test_get_timestamp_is_parsable(self) -> None:
        """Test that get_timestamp returns a parsable datetime string."""
        timestamp = get_timestamp()
        try:
            datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
            parsable = True
        except ValueError:
            parsable = False
        assert parsable

    def test_get_timestamp_format_matches_expected_pattern(self) -> None:
        """Test that get_timestamp matches the expected pattern."""
        timestamp = get_timestamp()
        assert len(timestamp) == 19  # YYYY-MM-DD HH:MM:SS (19 characters)

        # Check format with regex pattern
        pattern = r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}'
        assert re.match(pattern, timestamp) is not None



import json
import shutil
import traceback
from pathlib import Path
from typing import Any

from datamodel_code_generator import DataModelType, InputFileType, PythonVersion, generate
from jinja2 import Environment, FileSystemLoader

from .case import to_pascal_case, to_snake_case, to_title_case
from .logging import logger
from .specs import AppSpec, EndpointInfo
from .utils import get_source_dir, get_timestamp

TEMPLATES_DIR = get_source_dir() / '_internal' / 'templates'
"""Path to the templates directory."""

TEMPLATES_ENV = Environment(loader=FileSystemLoader(TEMPLATES_DIR), autoescape=True)
"""Jinja2 environment for rendering templates."""


def generate_mcp_file(
    output: Path,
    *,
    project_dir: Path | str | None,
    server_name: str,
    overwrite: bool = False,
) -> None:
    """Generate an MCP server file.

    Args:
        output: Path to the output file.
        project_dir: Path to the project directory.
        server_name: Name of the MCP server.
        overwrite: Whether to overwrite the output file if it exists.
    """
    logger.debug(f'Generating MCP server file to {output}')
    timestamp = get_timestamp()

    # Handle output
    if output.exists() and not overwrite:
        raise ValueError(f'Output path already exists: {output}')

    # Generate files
    template = TEMPLATES_ENV.get_template('server.j2')
    content = template.render(
        timestamp=timestamp,
        project_dir=project_dir or Path.cwd(),
        server_name=server_name,
    )
    content += '\n'
    output.write_text(content)

    logger.debug(f'Successfully generated MCP server file to {output}')


def generate_app_files(
    name: str,
    spec: AppSpec,
    base_dir: Path,
    *,
    overwrite: bool = False,
) -> None:
    """Generate files for the given application specification."""
    logger.debug(f'Generating files for {name!r}')
    timestamp = get_timestamp()

    # Handle output
    output_dir = base_dir / name
    if output_dir.exists() and not overwrite:
        raise ValueError(f'Output directory already exists: {output_dir}')
    shutil.rmtree(output_dir, ignore_errors=True)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Resolve template variables
    title = to_title_case(name)
    class_name = f'{to_pascal_case(name)}Client'

    # Generate files
    try:
        _generate_init_file(
            name,
            base_dir,
            uri=spec.uri,
            timestamp=timestamp,
            title=title,
            class_name=class_name,
        )

        _generate_client_file(
            name,
            base_dir,
            uri=spec.uri,
            timestamp=timestamp,
            title=title,
            class_name=class_name,
            base_url=spec.uri.rpartition('/')[0],
            endpoints=spec.lookup(),
        )

        _generate_models_file(
            name,
            base_dir,
            uri=spec.uri,
            timestamp=timestamp,
            title=title,
            data=spec.data,
        )

        logger.debug(f'Successfully generated files for {name!r}')

    except Exception as e:
        shutil.rmtree(output_dir, ignore_errors=True)
        raise e


def _generate_client_file(
    name: str,
    base_dir: Path,
    *,
    uri: str,
    timestamp: str,
    title: str,
    class_name: str,
    base_url: str,
    endpoints: list[EndpointInfo],
) -> None:
    """Generate client file."""
    logger.debug(f'Generating client file for {name!r}')
    output = base_dir / name / 'client.py'
    template = TEMPLATES_ENV.get_template('client.j2')
    content = template.render(
        name=name,
        uri=uri,
        timestamp=timestamp,
        title=title,
        class_name=class_name,
        endpoints=endpoints,
        base_url=base_url,
        get_query_params=_get_query_params,
        get_path_params=_get_path_params,
        get_python_type=_get_python_type,
        get_request_body_type=_get_request_body_type,
        get_response_model=_get_response_model,
        get_response_type=_get_response_type,
        has_query_params=_has_query_params,
        to_snake_case=to_snake_case,
    )
    content += '\n'
    output.write_text(content)


def _generate_init_file(
    name: str,
    base_dir: Path,
    *,
    uri: str,
    timestamp: str,
    title: str,
    class_name: str,
) -> None:
    """Generate init file."""
    logger.debug(f'Generating init file for {name!r}')
    output = base_dir / name / '__init__.py'
    template = TEMPLATES_ENV.get_template('__init__.j2')
    content = template.render(
        uri=uri,
        timestamp=timestamp,
        title=title,
        class_name=class_name,
    )
    content += '\n'
    output.write_text(content)


def _generate_models_file(
    name: str,
    base_dir: Path,
    *,
    uri: str,
    timestamp: str,
    title: str,
    data: dict[str, Any],
) -> None:
    """Generate models file using datamodel-code-generator."""
    logger.debug(f'Generating models file for {name!r}')
    output = base_dir / name / 'models.py'

    try:
        # Render models file header template
        template = TEMPLATES_ENV.get_template('models.j2')
        header = template.render(uri=uri, timestamp=timestamp, title=title)

        # Generate models using datamodel-code-generator
        generate(
            input_=json.dumps(data),
            input_file_type=InputFileType.OpenAPI,
            output=output,
            output_model_type=DataModelType.PydanticV2BaseModel,
            target_python_version=PythonVersion.PY_310,
            disable_appending_item_suffix=True,
            field_constraints=True,
            strict_nullable=True,
            use_annotated=True,
            use_schema_description=True,
            use_standard_collections=True,
            use_union_operator=True,
            custom_file_header=header,
        )

    except Exception as e:  # noqa: BLE001
        error = f'{e}\n{traceback.format_exc()}'
        logger.error(f'Error generating models file for {name!r}: {error}')
        template = TEMPLATES_ENV.get_template('models_error.j2')
        content = template.render(
            uri=uri,
            timestamp=timestamp,
            title=title,
            error=error,
        )
        content += '\n'
        output.write_text(content)


def _get_query_params(endpoint: EndpointInfo) -> list[dict[str, Any]]:
    """Get query parameters for the endpoint."""
    return [p for p in endpoint.parameters if p.get('in') == 'query']


def _get_path_params(endpoint: EndpointInfo) -> list[dict[str, Any]]:
    """Get path parameters for the endpoint."""
    return [p for p in endpoint.parameters if p.get('in') == 'path']


def _get_python_type(param: dict[str, Any]) -> str:
    """Convert OpenAPI parameter type to Python type."""
    mapping = {'integer': 'int', 'string': 'str', 'boolean': 'bool', 'number': 'float'}

    schema = param.get('schema', {})
    param_type = schema.get('type') if schema else param.get('type')
    if param_type in mapping:
        return mapping[param_type]
    return 'Any'


def _get_request_body_type(endpoint: EndpointInfo) -> str:
    """Get request body type."""
    assert endpoint.request_body, 'Endpoint must have a request body'
    schema = endpoint.request_body.get('content', {}).get('application/json', {}).get('schema', {})
    if schema.get('$ref'):
        name = schema.get('$ref').split('/')[-1]
        return f'models.{name}'
    return 'dict[str, Any]'


def _get_response_model(endpoint: EndpointInfo) -> str | None:
    """Extract the response model name from the endpoint response."""
    schema = (
        endpoint.responses.get('200', {})
        .get('content', {})
        .get('application/json', {})
        .get('schema', {})
    )
    if schema.get('$ref'):
        return f'models.{schema.get("$ref").split("/")[-1]}'
    return None


def _get_response_type(endpoint: EndpointInfo) -> str:
    """Convert OpenAPI response type to Python type."""
    return _get_response_model(endpoint) or 'dict[str, Any]'


def _has_query_params(endpoint: EndpointInfo) -> bool:
    """Check if the endpoint has query parameters."""
    return any(p.get('in') == 'query' for p in endpoint.parameters)



from pathlib import Path
from typing import NamedTuple

from mcp.server.fastmcp import FastMCP
from typing_extensions import Self

from .logging import logger
from .mcp import mount_to_fastmcp
from .registry import Registry
from .settings import AppSettings
from .specs import AppSpec, EndpointInfo, LookupConfig
from .store import Store
from .utils import get_project_dir

PROJECT_LOCK = 'apps.lock'
"""The file name for the project locked specifications."""

PROJECT_SETTINGS = 'apps.yaml'
"""The file name for the project settings."""


class ProjectPaths(NamedTuple):
    """Paths for the project directory and files."""

    dir: Path
    """The directory for the project settings and locked specifications files."""

    settings: Path
    """The file path for the project settings."""

    specs: Path
    """The file path for the project locked specifications."""


class Project:
    """Applications project.

    The project class manages a collection of applications defined by their settings and
    specifications. It provides methods to add, remove, and sync applications with a registry. The
    project maintains two primary files:
    - `apps.yaml`: A settings file that contains application settings in YAML format
    - `apps.lock`: A lock file that contains locked application specifications in JSON format
    """

    def __init__(
        self,
        project_dir: Path | str | None = None,
        registry_dir: Path | str | None = None,
    ) -> None:
        self.paths = _get_project_paths(project_dir)
        self.registry = Registry.load(registry_dir)
        self.specs = Store(self.paths.specs, AppSpec)
        self.settings = Store(self.paths.settings, AppSettings)

    async def __aenter__(self) -> Self:
        await self.registry.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        await self.registry.__aexit__(exc_type, exc_val, exc_tb)
        if exc_type is None:
            self.save()

    @classmethod
    def load(
        cls,
        project_dir: Path | str | None = None,
        registry_dir: Path | str | None = None,
    ) -> Self:
        """Load the applications project from the given project and registry directories."""
        self = cls(project_dir, registry_dir)
        self.settings.read()
        self.specs.read()
        return self

    def save(self) -> None:
        """Save the applications project settings and locked specifications files."""
        self.settings.write()
        self.specs.write()

    async def add_app(
        self,
        name: str,
        settings: AppSettings,
        *,
        force: bool = False,
        frozen: bool = False,
        upgrade: bool = False,
        sync: bool = False,
    ) -> None:
        """Add an application to the project.

        Args:
            name: The name of the application
            settings: The application settings
            force: Whether to overwrite application if it already exists with a different URI
            frozen: Whether to skip adding application specification in the project lock file
            upgrade: Whether to force upgrading application specification
            sync: Whether to sync application specification with the project registry
        """
        logger.info(f'Adding {name!r} application to project...')
        # Handle existing application
        if name in self.settings:
            current_settings = self.settings[name]
            if settings.uri != current_settings.uri:
                if not force:
                    raise ValueError(f'Project application {name!r} exists with different URI')
                logger.warning(f'Upgrading project application {name!r} URI to {settings.uri}')
        # Add application specification
        if not frozen:
            current_spec = self.specs.get(name)
            if upgrade or current_spec is None or current_spec.uri != settings.uri:
                try:
                    self.specs[name] = await AppSpec.fetch(settings.uri)
                except Exception as e:
                    raise ValueError(f'Failed to fetch specification for {name!r}') from e
            if sync:
                self.registry.add_app(
                    name,
                    self.specs[name],
                    force=force,
                    upgrade=upgrade,
                )
        # Add application settings
        self.settings[name] = settings

    async def remove_app(
        self,
        name: str,
        *,
        frozen: bool = False,
        sync: bool = False,
    ) -> None:
        """Remove an application from the project.

        Args:
            name: The name of the application
            frozen: Whether to skip removing application specification from the project lock file
            sync: Whether to sync application specification with the project registry
        """
        logger.info(f'Removing {name!r} application from project...')
        # Handle non-existing application
        if name not in self.settings:
            logger.warning(f'Project application {name!r} is not registered')
            return
        # Remove application specification
        if not frozen:
            if sync:
                try:
                    self.registry.remove_app(name)
                except Exception as e:  # noqa: BLE001
                    logger.warning(f'Failed to remove application {name!r} from registry: {e}')
            if name in self.specs:
                del self.specs[name]
        # Remove application settings
        del self.settings[name]

    async def lock(
        self,
        *,
        force: bool = False,
        upgrade: bool = False,
    ) -> None:
        """Lock project specifications.

        Args:
            force: Whether to overwrite application if it already exists with a different URI
            upgrade: Whether to force upgrading application specification
        """
        logger.info('Locking project specifications...')
        specs = {name: spec for name, spec in self.specs.items() if name in self.settings}
        for name, settings in self.settings.items():
            if name in specs:
                spec = specs[name]
                if spec.uri != settings.uri:
                    if not force:
                        raise ValueError(f'Project application {name!r} exists with different URI')
                    logger.warning(f'Upgrading project application {name!r} URI to {settings.uri}')
                elif upgrade:
                    logger.info(f'Upgrading project application {name!r}')
                else:
                    logger.info(f'Project application {name!r} requirements already satisfied')
                    continue
            try:
                specs[name] = await AppSpec.fetch(settings.uri)
            except Exception as e:
                raise ValueError(f'Failed to fetch specification for {name!r}') from e
        self.specs.clear()
        self.specs.update(specs)

    async def sync(
        self,
        *,
        force: bool = False,
        frozen: bool = False,
        upgrade: bool = False,
    ) -> None:
        """Sync applications with the project registry.

        Args:
            force: Whether to overwrite application if it already exists with a different URI
            frozen: Whether to skip updating application specification in the project lock file
            upgrade: Whether to force upgrading application specification
        """
        logger.info('Syncing applications with project registry...')
        # Lock project specifications
        if not frozen:
            try:
                await self.lock(force=force, upgrade=upgrade)
            except Exception as e:
                raise ValueError('Failed to lock project specifications') from e
        # Sync applications with the project registry
        for name, spec in self.specs.items():
            try:
                self.registry.add_app(
                    name,
                    spec,
                    force=force,
                    upgrade=upgrade,
                )
            except Exception as e:  # noqa: BLE001
                raise ValueError(f'Failed to sync application {name!r}: {e}') from e

    def lookup(self, **config: LookupConfig | bool | None) -> dict[str, list[EndpointInfo]]:
        """Lookup endpoints with the given application lookup configurations.

        Args:
            **config: Application lookup configuration or boolean to include all applications
                endpoints. When provided, only applications matching the filters will be included.
                Otherwise, all applications endpoints will be included.

        Returns:
            Dictionary mapping application names to lists of endpoint information.
        """
        endpoints = {}
        for name, settings in self.settings.items():
            spec = self.specs[name]
            if spec is None:
                logger.warning(f'Project application {name!r} specification is missing')
                continue
            # Resolve lookup configuration
            if config and name not in config:
                continue
            lookup_config = config.get(name)
            if lookup_config is None:
                lookup_config = settings.get_lookup_config()
            elif lookup_config is False:
                continue
            elif lookup_config is True:
                lookup_config = {}
            # Apply lookup configuration
            endpoints[name] = spec.lookup(**lookup_config)
        return endpoints

    def mount(self, __server: FastMCP, **config: LookupConfig | bool | None) -> None:
        """Mount the project to a FastMCP server with the given lookup configuration.

        Args:
            server: The FastMCP server to mount the project to
            **config: Application lookup configuration or boolean to include all applications
                endpoints. When provided, only applications matching the filters will be included.
                Otherwise, all applications endpoints will be included.
        """
        endpoints = self.lookup(**config)
        mount_to_fastmcp(__server, endpoints)


def _get_project_paths(project_dir: Path | str | None = None) -> ProjectPaths:
    """Get the project directory and file paths."""
    project_dir = Path(project_dir) if project_dir else get_project_dir()
    return ProjectPaths(
        dir=project_dir,
        settings=project_dir / PROJECT_SETTINGS,
        specs=project_dir / PROJECT_LOCK,
    )



import functools
import typing
from typing import Any

from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

from .case import to_pascal_case, to_snake_case

if typing.TYPE_CHECKING:
    from .clients import ClientConfig
    from .specs import LookupConfig


class AppSettings(BaseSettings):
    """Application settings."""

    model_config = SettingsConfigDict(extra='allow')

    uri: str = Field(..., frozen=True)
    """URI to OpenAPI specification."""

    base_url: str | None = None
    """Base URL for API."""

    include_methods: list[str] | None = None
    """Methods to include in mounted endpoints."""

    exclude_methods: list[str] | None = None
    """Methods to exclude from mounted endpoints."""

    include_tags: list[str] | None = None
    """Tags to include in mounted endpoints."""

    exclude_tags: list[str] | None = None
    """Tags to exclude from mounted endpoints."""

    filters: dict[str, str] | None = None
    """Filters to apply to mounted endpoints."""

    @model_validator(mode='after')
    def model_validate_base_url(self) -> Self:
        if not self.base_url:
            self.base_url = self.uri.rpartition('/')[0]
        return self

    def get_client_config(self) -> 'ClientConfig':
        """Get client configuration."""
        return self.model_dump(  # type: ignore
            include={
                'base_url',
                'api_key',
                'token',
                'api_key_header',
                'token_header',
                'timeout',
            },
        )

    def get_lookup_config(self) -> 'LookupConfig':
        """Get endpoint lookup configuration."""
        return self.model_dump(  # type: ignore
            include={
                'include_methods',
                'exclude_methods',
                'include_tags',
                'exclude_tags',
                'filters',
            },
        )


class ClientSettings(BaseSettings):
    """Client settings."""

    model_config = SettingsConfigDict(extra='allow')

    name: str = Field(..., frozen=True)
    """Application name."""

    base_url: str
    """Base URL for API."""

    api_key: str | None = None
    """API key for authentication."""

    token: str | None = None
    """Bearer token for authentication."""

    api_key_header: str = 'X-API-Key'
    """Header name for API key."""

    token_header: str = 'Authorization'  # noqa: S105
    """Header name for bearer token."""

    timeout: float = 30
    """Request timeout in seconds."""

    @classmethod
    def with_env_prefix(cls, **data: Any) -> Self:
        """Create a new instance with client-specific environment variable prefix."""
        # Create a new settings instance that loads values from environment variables with the
        # application prefix "APPDOG_{APP_NAME}_" to allow for client-specific configuration.
        assert 'name' in data, 'Client name is required'
        env_prefix = f'APPDOG_{to_snake_case(data["name"]).upper()}_'
        cls = type(
            f'{to_pascal_case(data["name"])}Settings',
            (cls,),
            {'model_config': SettingsConfigDict(env_prefix=env_prefix, extra='allow')},
        )
        return cls(**data)

    def get_headers(self) -> dict[str, str]:
        """Get headers for client requests."""
        return _compute_headers(self.api_key, self.token, self.api_key_header, self.token_header)


@functools.lru_cache(maxsize=1)
def _compute_headers(
    api_key: str | None,
    token: str | None,
    api_key_header: str,
    token_header: str,
) -> dict[str, str]:
    """Compute headers with caching based on auth credentials."""
    headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
    }
    if api_key:
        headers[api_key_header] = api_key
    if token:
        headers[token_header] = f'Bearer {token}'
    return headers



import os
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from appdog._internal.settings import AppSettings, ClientSettings, _compute_headers


class TestAppSettings:
    """Tests for AppSettings class."""

    def test_init_with_required_fields(self) -> None:
        """Test initialization with required fields."""
        settings = AppSettings(uri='http://example.com/api')
        assert settings.uri == 'http://example.com/api'
        assert settings.base_url == 'http://example.com'
        assert settings.include_methods is None
        assert settings.exclude_methods is None
        assert settings.include_tags is None
        assert settings.exclude_tags is None

    def test_init_with_optional_fields(self) -> None:
        """Test initialization with optional fields."""
        settings = AppSettings(
            uri='http://example.com/api',
            base_url='http://api.example.com',
            include_methods=['GET', 'POST'],
            exclude_methods=['DELETE'],
            include_tags=['public'],
            exclude_tags=['private'],
        )
        assert settings.uri == 'http://example.com/api'
        assert settings.base_url == 'http://api.example.com'
        assert settings.include_methods == ['GET', 'POST']
        assert settings.exclude_methods == ['DELETE']
        assert settings.include_tags == ['public']
        assert settings.exclude_tags == ['private']

    def test_init_without_uri(self) -> None:
        """Test initialization without required uri field."""
        with pytest.raises(ValidationError):
            AppSettings()  # type: ignore

    def test_model_validate_base_url(self) -> None:
        """Test base_url validation."""
        settings = AppSettings(uri='https://api.example.com/v1')
        assert settings.base_url == 'https://api.example.com'

        settings = AppSettings(uri='https://api.example.com/v1', base_url='https://api.example.com')
        assert settings.base_url == 'https://api.example.com'

    def test_extra_fields_allowed(self) -> None:
        """Test that extra fields are allowed."""
        settings = AppSettings(uri='http://example.com/api', extra_field='value')  # type: ignore
        assert hasattr(settings, 'extra_field')
        assert settings.extra_field == 'value'

    def test_get_client_config(self) -> None:
        """Test get_client_config method."""
        settings = AppSettings(  # type: ignore
            uri='http://example.com/api',
            base_url='http://api.example.com',
            api_key='test-key',
            token='test-token',  # noqa: S106
            api_key_header='X-API-Key',
            token_header='Auth',  # noqa: S106
            timeout=60,
            include_methods=['GET'],  # Should be excluded from client config
        )

        config = settings.get_client_config()
        assert config['base_url'] == 'http://api.example.com'
        assert config['api_key'] == 'test-key'
        assert config['token'] == 'test-token'  # noqa: S105
        assert config['api_key_header'] == 'X-API-Key'
        assert config['token_header'] == 'Auth'  # noqa: S105
        assert config['timeout'] == 60
        assert 'include_methods' not in config

    def test_get_lookup_config(self) -> None:
        """Test get_lookup_config method."""
        settings = AppSettings(  # type: ignore
            uri='http://example.com/api',
            include_methods=['GET', 'POST'],
            exclude_methods=['DELETE'],
            include_tags=['public'],
            exclude_tags=['private'],
            filters={'path': '/api/v1/.*'},
            api_key='test-key',  # Should be excluded from lookup config
        )

        config = settings.get_lookup_config()
        assert config['include_methods'] == ['GET', 'POST']
        assert config['exclude_methods'] == ['DELETE']
        assert config['include_tags'] == ['public']
        assert config['exclude_tags'] == ['private']
        assert config['filters'] == {'path': '/api/v1/.*'}
        assert 'api_key' not in config
        assert 'uri' not in config


class TestClientSettings:
    """Tests for ClientSettings class."""

    def test_init_with_required_fields(self) -> None:
        """Test initialization with required fields."""
        settings = ClientSettings(name='TestApp', base_url='http://example.com')
        assert settings.name == 'TestApp'
        assert settings.base_url == 'http://example.com'
        assert settings.api_key is None
        assert settings.token is None
        assert settings.api_key_header == 'X-API-Key'
        assert settings.token_header == 'Authorization'  # noqa: S105
        assert settings.timeout == 30

    def test_init_with_optional_fields(self) -> None:
        """Test initialization with optional fields."""
        settings = ClientSettings(
            name='TestApp',
            base_url='http://example.com',
            api_key='test-key',
            token='test-token',  # noqa: S106
            api_key_header='X-Custom-Key',
            token_header='X-Custom-Token',  # noqa: S106
            timeout=60,
        )
        assert settings.name == 'TestApp'
        assert settings.base_url == 'http://example.com'
        assert settings.api_key == 'test-key'
        assert settings.token == 'test-token'  # noqa: S105
        assert settings.api_key_header == 'X-Custom-Key'
        assert settings.token_header == 'X-Custom-Token'  # noqa: S105
        assert settings.timeout == 60

    def test_init_without_required_fields(self) -> None:
        """Test initialization without required fields."""
        with pytest.raises(ValidationError):
            ClientSettings()  # type: ignore

        with pytest.raises(ValidationError):
            ClientSettings(name='TestApp')  # type: ignore

        with pytest.raises(ValidationError):
            ClientSettings(base_url='http://example.com')  # type: ignore

    def test_with_env_prefix(self) -> None:
        """Test creating settings with environment variable prefix."""
        with patch.dict(
            os.environ,
            {
                'APPDOG_TEST_APP_API_KEY': 'env-key',
                'APPDOG_TEST_APP_TOKEN': 'env-token',
                'APPDOG_TEST_APP_TIMEOUT': '45',
            },
        ):
            settings = ClientSettings.with_env_prefix(
                name='TestApp',
                base_url='http://example.com',
            )
            assert settings.name == 'TestApp'
            assert settings.base_url == 'http://example.com'
            assert settings.api_key == 'env-key'
            assert settings.token == 'env-token'  # noqa: S105
            assert settings.timeout == 45

    def test_get_headers_no_auth(self) -> None:
        """Test getting headers without authentication."""
        settings = ClientSettings(name='TestApp', base_url='http://example.com')
        headers = settings.get_headers()
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }

    def test_get_headers_with_api_key(self) -> None:
        """Test getting headers with API key authentication."""
        settings = ClientSettings(
            name='TestApp',
            base_url='http://example.com',
            api_key='test-key',
        )
        headers = settings.get_headers()
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-API-Key': 'test-key',
        }

    def test_get_headers_with_token(self) -> None:
        """Test getting headers with token authentication."""
        settings = ClientSettings(
            name='TestApp',
            base_url='http://example.com',
            token='test-token',  # noqa: S106
        )
        headers = settings.get_headers()
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer test-token',
        }

    def test_get_headers_with_both_auth(self) -> None:
        """Test getting headers with both API key and token authentication."""
        settings = ClientSettings(
            name='TestApp',
            base_url='http://example.com',
            api_key='test-key',
            token='test-token',  # noqa: S106
        )
        headers = settings.get_headers()
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-API-Key': 'test-key',
            'Authorization': 'Bearer test-token',
        }

    def test_extra_fields_allowed(self) -> None:
        """Test that extra fields are allowed."""
        settings = ClientSettings(  # type: ignore
            name='TestApp',
            base_url='http://example.com',
            extra_field='value',
        )
        assert hasattr(settings, 'extra_field')
        assert settings.extra_field == 'value'


class TestComputeHeaders:
    """Tests for _compute_headers function."""

    def test_compute_headers_no_auth(self) -> None:
        """Test computing headers without authentication."""
        headers = _compute_headers(None, None, 'X-API-Key', 'Authorization')
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }

    def test_compute_headers_with_api_key(self) -> None:
        """Test computing headers with API key authentication."""
        headers = _compute_headers('test-key', None, 'X-API-Key', 'Authorization')
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-API-Key': 'test-key',
        }

    def test_compute_headers_with_token(self) -> None:
        """Test computing headers with token authentication."""
        headers = _compute_headers(None, 'test-token', 'X-API-Key', 'Authorization')
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer test-token',
        }

    def test_compute_headers_with_both_auth(self) -> None:
        """Test computing headers with both API key and token authentication."""
        headers = _compute_headers('test-key', 'test-token', 'X-API-Key', 'Authorization')
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-API-Key': 'test-key',
            'Authorization': 'Bearer test-token',
        }

    def test_compute_headers_custom_headers(self) -> None:
        """Test computing headers with custom header names."""
        headers = _compute_headers(
            'test-key',
            'test-token',
            'X-Custom-Key',
            'X-Custom-Token',
        )
        assert headers == {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'X-Custom-Key': 'test-key',
            'X-Custom-Token': 'Bearer test-token',
        }

    def test_compute_headers_caching(self) -> None:
        """Test that _compute_headers is cached."""
        headers1 = _compute_headers('test-key', 'test-token', 'X-API-Key', 'Authorization')
        headers2 = _compute_headers('test-key', 'test-token', 'X-API-Key', 'Authorization')
        assert headers1 is headers2  # Same object due to caching



import json
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from typing_extensions import Self

from tests.conftest import AsyncNoWarningMock


class TestImports:
    """Tests for module imports in the appdog package."""

    @mock.patch('builtins.__import__')
    def test_appdog_petstore_import(self, mock_import: mock.MagicMock) -> None:
        """Test that appdog.petstore can be imported."""
        # Create a mock module
        mock_module = mock.MagicMock()
        mock_module.PetstoreClient = mock.MagicMock()
        mock_module.client = mock.MagicMock()
        mock_module.models = mock.MagicMock()

        # Configure import to return our mock
        mock_import.return_value = mock_module

        # Try importing
        try:
            # This will use our mocked __import__
            module = __import__('appdog.petstore', fromlist=['*'])

            # Verify it worked
            assert module is mock_module
            assert hasattr(module, 'PetstoreClient')
            assert hasattr(module, 'client')
            assert hasattr(module, 'models')
        except ImportError as e:
            pytest.fail(f'Failed to import appdog.petstore: {e}')

        # Verify import was called correctly
        mock_import.assert_called_with('appdog.petstore', fromlist=['*'])

    @mock.patch('builtins.__import__')
    def test_petstore_client_attributes(self, mock_import: mock.MagicMock) -> None:
        """Test that the PetstoreClient class has expected attributes and methods."""
        # Create mock client class
        mock_client_class = mock.MagicMock()
        mock_client_class.__aenter__ = AsyncNoWarningMock()
        mock_client_class.__aexit__ = AsyncNoWarningMock()
        mock_client_class.get_pet_find_by_status = AsyncNoWarningMock()

        # Create mock client instance
        mock_client_instance = mock.MagicMock()

        # Create a mock module
        mock_module = mock.MagicMock()
        mock_module.PetstoreClient = mock_client_class
        mock_module.client = mock_client_instance

        # Configure import to return our mock
        mock_import.return_value = mock_module

        # Import the module
        module = __import__('appdog.petstore', fromlist=['*'])

        # Check PetstoreClient class exists
        assert hasattr(module, 'PetstoreClient')

        # Check client instance exists
        assert hasattr(module, 'client')

        # Check expected client methods
        client_class = module.PetstoreClient
        assert hasattr(client_class, 'get_pet_find_by_status')
        assert hasattr(client_class, '__aenter__')
        assert hasattr(client_class, '__aexit__')

    @mock.patch('builtins.__import__')
    def test_petstore_models_import(self, mock_import: mock.MagicMock) -> None:
        """Test that appdog.petstore.models can be imported and has expected classes."""
        # Create a mock models module
        mock_models = mock.MagicMock()

        # Create a mock petstore module
        mock_module = mock.MagicMock()
        mock_module.models = mock_models

        # Configure import to return our mock
        mock_import.return_value = mock_module

        try:
            # Import the module
            module = __import__('appdog.petstore.models', fromlist=['*'])

            # Verify it worked
            assert isinstance(module, mock.MagicMock)
        except ImportError as e:
            pytest.fail(f'Failed to import appdog.petstore.models: {e}')


@pytest.fixture
def petstore_response() -> dict[str, Any]:
    """Load the petstore response fixture."""
    fixtures_path = Path(__file__).parent / 'fixtures'
    with open(fixtures_path / 'response_petstore.json') as f:
        return json.load(f)  # type: ignore


@pytest.mark.asyncio
class TestPetstoreAsyncClient:
    """Tests for the async client functionality."""

    @mock.patch('builtins.__import__')
    async def test_petstore_client_context_manager(
        self, mock_import: mock.MagicMock, monkeypatch: Any, petstore_response: dict[str, Any]
    ) -> None:
        """Test that the PetstoreClient can be used as an async context manager."""
        # Create a mock client class with async methods
        mock_get_pet = AsyncNoWarningMock(return_value=petstore_response)

        # Create a mock client instance
        class MockClient:
            get_pet_find_by_status = mock_get_pet

            async def __aenter__(self) -> Self:
                return self

            async def __aexit__(
                self,
                exc_type: type[BaseException] | None,
                exc_val: BaseException | None,
                exc_tb: object,
            ) -> None:
                pass

        mock_client_instance = MockClient()

        # Create a mock module
        mock_module = mock.MagicMock()
        mock_module.client = mock_client_instance

        # Configure import to return our mock
        mock_import.return_value = mock_module

        # Import the module
        module = __import__('appdog.petstore', fromlist=['*'])

        # Test using the client as context manager
        async with module.client as client:
            result = await client.get_pet_find_by_status(status='available')

        # Check the mock was called correctly
        mock_get_pet.assert_called_once()
        assert result == petstore_response
        assert result['id'] == 1
        assert result['name'] == 'doggie'
        assert result['status'] == 'available'
